# Generated by Django 4.1.8 on 2023-04-28 14:57
import uuid

import django.core.serializers.json
import django.db.models.deletion
import wbcore.contrib.icons.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("directory", "0004_entry_is_draft_entry"),
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("authentication", "0005_user_external_calendar_settings"),
    ]

    operations = [
        migrations.CreateModel(
            name="Workflow",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, unique=True, verbose_name="Name")),
                (
                    "single_instance_execution",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates wether the workflow is restricted to only one running process per workflow or attached model instance",
                        verbose_name="Single Instance Execution",
                    ),
                ),
                (
                    "status_field",
                    models.CharField(
                        blank=True,
                        help_text="The status field name in the corresponding model that needs to be updated by the workflow",
                        max_length=64,
                        null=True,
                        verbose_name="Status Field",
                    ),
                ),
                (
                    "model",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contenttypes.contenttype",
                        verbose_name="Model",
                    ),
                ),
                (
                    "graph",
                    models.ImageField(blank=True, null=True, upload_to="", verbose_name="Graph"),
                ),
            ],
            options={
                "verbose_name": "Workflow",
                "verbose_name_plural": "Workflows",
            },
        ),
        migrations.AddField(
            model_name="workflow",
            name="preserve_instance",
            field=models.BooleanField(
                default=False,
                help_text="This will preserve the attached instance's state in a JSON field when starting the workflow. You then have the option to write this data back onto the instance in a finish step, effectively preserving the instance's state throughout the workflow.",
                verbose_name="Preserve Instance",
            ),
        ),
        migrations.CreateModel(
            name="Step",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, verbose_name="Name")),
                ("code", models.PositiveIntegerField(verbose_name="Code")),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        help_text="The status that will be set in the attached model upon transitioning to this step",
                        max_length=64,
                        null=True,
                        verbose_name="Status",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="workflow.workflow",
                        verbose_name="Workflow",
                        related_name="associated_steps",
                    ),
                ),
                (
                    "step_type",
                    models.CharField(
                        choices=[
                            ("StartStep", "Start Step"),
                            ("UserStep", "User Step"),
                            ("DecisionStep", "Decision Step"),
                            ("SplitStep", "Split Step"),
                            ("JoinStep", "Join Step"),
                            ("EmailStep", "Email Step"),
                            ("ScriptStep", "Script Step"),
                            ("FinishStep", "Finish Step"),
                        ],
                        default="",
                        max_length=64,
                        verbose_name="Step Type",
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        blank=True,
                        help_text="Define which permission is needed to be able to view this step being executed.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="related_steps",
                        to="authentication.permission",
                        verbose_name="Permission",
                    ),
                ),
            ],
            options={
                "verbose_name": "Step",
                "verbose_name_plural": "Steps",
            },
        ),
        migrations.AddConstraint(
            model_name="step",
            constraint=models.UniqueConstraint(fields=("name", "workflow"), name="unique_name_workflow"),
        ),
        migrations.AddConstraint(
            model_name="step",
            constraint=models.UniqueConstraint(fields=("code", "workflow"), name="unique_code_workflow"),
        ),
        migrations.CreateModel(
            name="Transition",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, verbose_name="Name")),
            ],
            options={
                "verbose_name": "Transition",
                "verbose_name_plural": "Transitions",
            },
        ),
        migrations.AddField(
            model_name="transition",
            name="from_step",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="outgoing_transitions",
                to="workflow.step",
                verbose_name="From",
            ),
        ),
        migrations.AddField(
            model_name="transition",
            name="to_step",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="incoming_transitions",
                to="workflow.step",
                verbose_name="To",
            ),
        ),
        migrations.AddField(
            model_name="transition",
            name="icon",
            field=wbcore.contrib.icons.models.IconField(default="start", max_length=128, verbose_name="Icon"),
        ),
        migrations.AddConstraint(
            model_name="transition",
            constraint=models.UniqueConstraint(fields=("name", "to_step"), name="unique_name_to_step"),
        ),
        migrations.CreateModel(
            name="Condition",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "attribute_name",
                    models.CharField(
                        help_text="The name of the workflow data or attached model field whose value should be checked",
                        max_length=64,
                        verbose_name="Attribute Name",
                    ),
                ),
                (
                    "operator",
                    models.CharField(
                        choices=[
                            ("<", "Less Than"),
                            ("<=", "Less Than or Equal"),
                            ("==", "Equal"),
                            (">=", "Greater Than or Equal"),
                            (">", "Greater Than"),
                        ],
                        max_length=64,
                        verbose_name="Operator",
                    ),
                ),
                (
                    "negate_operator",
                    models.BooleanField(default=False, verbose_name="Negate Operator"),
                ),
                (
                    "expected_value",
                    models.CharField(
                        help_text="The expected attribute's value",
                        max_length=64,
                        verbose_name="Expected Value",
                    ),
                ),
                (
                    "transition",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="associated_conditions",
                        to="workflow.transition",
                        verbose_name="Transition",
                    ),
                ),
            ],
            options={
                "verbose_name": "Condition",
                "verbose_name_plural": "Conditions",
            },
        ),
        migrations.AddConstraint(
            model_name="condition",
            constraint=models.UniqueConstraint(
                fields=("transition", "attribute_name", "operator", "negate_operator", "expected_value"),
                name="unique_transition_attribute_name_operator_negate_expected_value",
            ),
        ),
        migrations.CreateModel(
            name="StartStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
            ],
            options={
                "verbose_name": "Start Step",
                "verbose_name_plural": "Start Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="UserStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
                ("notify_user", models.BooleanField(default=True, verbose_name="Notify User")),
                (
                    "assignee",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="assigned_workflow_steps",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Assignee",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="assigned_workflow_steps",
                        to="authentication.group",
                        verbose_name="Group",
                    ),
                ),
            ],
            options={
                "verbose_name": "User Step",
                "verbose_name_plural": "User Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.AddField(
            model_name="userstep",
            name="kwargs",
            field=models.JSONField(blank=True, null=True, verbose_name="Kwargs"),
        ),
        migrations.AddField(
            model_name="userstep",
            name="assignee_method",
            field=models.CharField(
                blank=True,
                choices=[
                    ("manager_of_instance_assignee", "Instance Assignee's Manager"),
                    ("weighted_random", "Weighted Random Group Member"),
                    ("random_group_member", "Random Group Member"),
                    ("activity_assignee", "Activity Assignee"),
                ],
                max_length=64,
                null=True,
                verbose_name="Assignee Method",
            ),
        ),
        migrations.AddConstraint(
            model_name="userstep",
            constraint=models.CheckConstraint(
                check=models.Q(("assignee__isnull", False), ("group__isnull", False), _negated=True),
                name="check_not_both_assignee_group",
            ),
        ),
        migrations.CreateModel(
            name="Process",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name="UUID"
                    ),
                ),
                ("started", models.DateTimeField(auto_now_add=True, verbose_name="Started")),
                ("finished", models.DateTimeField(blank=True, null=True, verbose_name="Finished", editable=False)),
                (
                    "instance_id",
                    models.CharField(blank=True, editable=False, max_length=128, null=True, verbose_name="Instance"),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contenttypes.contenttype",
                        verbose_name="Content Type",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="workflow.workflow",
                        verbose_name="Workflow",
                        editable=False,
                    ),
                ),
            ],
            options={
                "verbose_name": "Process",
                "verbose_name_plural": "Processes",
            },
        ),
        migrations.AddField(
            model_name="process",
            name="preserved_instance",
            field=models.JSONField(
                blank=True,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
                verbose_name="Preserved Instance",
            ),
        ),
        migrations.AddField(
            model_name="process",
            name="state",
            field=models.CharField(
                choices=[("Active", "Active"), ("Finished", "Finished"), ("Failed", "Failed")],
                default="Active",
                editable=False,
                help_text="The current state of this process. Can be one of: 'Active', 'Finished' or 'Failed'",
                max_length=64,
                verbose_name="State",
            ),
        ),
        migrations.CreateModel(
            name="ProcessStep",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name="UUID"
                    ),
                ),
                ("started", models.DateTimeField(auto_now_add=True, verbose_name="Started")),
                ("finished", models.DateTimeField(blank=True, editable=False, null=True, verbose_name="Finished")),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("Active", "Active"),
                            ("Finished", "Finished"),
                            ("Waiting", "Waiting"),
                            ("Failed", "Failed"),
                            ("Canceled", "Canceled"),
                        ],
                        editable=False,
                        help_text="The current state of this process step. Can be one of: 'Active', 'Finished', 'Waiting', 'Failed' or 'Canceled'",
                        max_length=64,
                        verbose_name="State",
                        default="Active",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="The status that will be set in the attached model upon transitioning to this step",
                        max_length=64,
                        null=True,
                        verbose_name="Status",
                    ),
                ),
                (
                    "assignee",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="assigned_process_steps",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Assignee",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="assigned_process_steps",
                        to="authentication.group",
                        verbose_name="Group",
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="related_process_steps",
                        to="authentication.permission",
                        verbose_name="Permission",
                        editable=False,
                        help_text="The permission needed to be able to view this step being executed.",
                    ),
                ),
                (
                    "process",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="process_steps",
                        to="workflow.process",
                        verbose_name="Process",
                    ),
                ),
                (
                    "step",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="process_steps",
                        to="workflow.step",
                        verbose_name="Step",
                    ),
                ),
                (
                    "error_message",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="An error message will be displayed when this step fails",
                        max_length=128,
                        null=True,
                        verbose_name="Error Message",
                    ),
                ),
            ],
            options={"verbose_name": "Process Step", "verbose_name_plural": "Process Steps"},
        ),
        migrations.CreateModel(
            name="Display",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("grid_template_areas", models.JSONField(verbose_name="Grid Fields")),
                ("name", models.CharField(max_length=128, unique=True, verbose_name="Name")),
            ],
            options={
                "verbose_name": "Display",
                "verbose_name_plural": "Displays",
            },
        ),
        migrations.AddField(
            model_name="userstep",
            name="display",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="user_steps",
                to="workflow.display",
                verbose_name="Display",
            ),
        ),
        migrations.CreateModel(
            name="DecisionStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
            ],
            options={"verbose_name": "Decision Step", "verbose_name_plural": "Decision Steps"},
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="SplitStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
            ],
            options={
                "verbose_name": "Split Step",
                "verbose_name_plural": "Split Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="FinishStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
            ],
            options={
                "verbose_name": "Finish Step",
                "verbose_name_plural": "Finish Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.AddField(
            model_name="finishstep",
            name="write_preserved_instance",
            field=models.BooleanField(
                default=False,
                help_text="Writes the instance data backup onto the instance restoring its state to the beginning of the workflow.",
                verbose_name="Write Preserved Instance",
            ),
        ),
        migrations.AlterField(
            model_name="step",
            name="step_type",
            field=models.CharField(
                choices=[
                    ("UserStep", "User Step"),
                    ("DecisionStep", "Decision Step"),
                    ("SplitStep", "Split Step"),
                    ("JoinStep", "Join Step"),
                    ("EmailStep", "Email Step"),
                    ("ScriptStep", "Script Step"),
                    ("FinishStep", "Finish Step"),
                ],
                default="",
                max_length=64,
                verbose_name="Step Type",
            ),
        ),
        migrations.CreateModel(
            name="JoinStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
                (
                    "wait_for_all",
                    models.BooleanField(
                        default=True,
                        help_text="If False, will cancel all other incoming process steps as soon the first reaches this step.",
                        verbose_name="Wait For All",
                    ),
                ),
            ],
            options={
                "verbose_name": "Join Step",
                "verbose_name_plural": "Join Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="ScriptStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
                ("script", models.TextField(verbose_name="Script")),
            ],
            options={
                "verbose_name": "Script Step",
                "verbose_name_plural": "Script Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="EmailStep",
            fields=[
                (
                    "step_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflow.step",
                    ),
                ),
                (
                    "template",
                    models.FileField(
                        max_length=256, upload_to="workflow/templates/Test_Templates.txt", verbose_name="Template"
                    ),
                ),
                ("subject", models.CharField(max_length=128, verbose_name="Subject")),
                (
                    "to",
                    models.ManyToManyField(
                        related_name="workflow_step_receiver", to="directory.emailcontact", verbose_name="To"
                    ),
                ),
                (
                    "cc",
                    models.ManyToManyField(
                        related_name="workflow_steps_cc", to="directory.emailcontact", verbose_name="CC", blank=True
                    ),
                ),
                (
                    "bcc",
                    models.ManyToManyField(
                        related_name="workflow_steps_bcc", to="directory.emailcontact", verbose_name="BCC", blank=True
                    ),
                ),
            ],
            options={
                "verbose_name": "Email Step",
                "verbose_name_plural": "Email Steps",
            },
            bases=("workflow.step",),
        ),
        migrations.CreateModel(
            name="Data",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("label", models.CharField(max_length=64, verbose_name="Label")),
                ("help_text", models.CharField(blank=True, max_length=128, null=True, verbose_name="Help Text")),
                (
                    "data_type",
                    models.CharField(
                        choices=[
                            ("Text", "Text"),
                            ("Number", "Number"),
                            ("Date", "Date"),
                            ("Date Time", "Date Time"),
                            ("Boolean", "Boolean"),
                        ],
                        default="Text",
                        max_length=64,
                        verbose_name="Data Type",
                    ),
                ),
                ("required", models.BooleanField(verbose_name="Required")),
                ("default", models.CharField(max_length=128, verbose_name="Default", blank=True, null=True)),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attached_data",
                        to="workflow.workflow",
                        verbose_name="Workflow",
                    ),
                ),
            ],
            options={
                "verbose_name": "Data",
                "verbose_name_plural": "Data",
            },
        ),
        migrations.AddConstraint(
            model_name="data",
            constraint=models.UniqueConstraint(fields=("label", "workflow"), name="unique_label_workflow"),
        ),
        migrations.CreateModel(
            name="DataValue",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("value", models.CharField(max_length=64, verbose_name="Value", blank=True, null=True)),
                (
                    "data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="workflow.data",
                        verbose_name="Data",
                    ),
                ),
                (
                    "process",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data_values",
                        to="workflow.process",
                        verbose_name="Process",
                    ),
                ),
            ],
            options={
                "verbose_name": "Data Value",
                "verbose_name_plural": "Data Values",
            },
        ),
    ]



from wbcore.contrib.icons.serializers import IconSelectField

from .boolean import BooleanField
from .choice import ChoiceField, MultipleChoiceField
from .datetime import (
    DateField,
    DateRangeField,
    DateTimeField,
    DateTimeRangeField,
    DurationField,
    TimeField,
    TimeZoneField,
)
from .fields import (
    AdditionalResourcesField,
    DynamicButtonField,
    HyperlinkField,
    ReadOnlyField,
    SlugRelatedField,
    SerializerMethodField,
    register_dynamic_button,
    register_only_instance_dynamic_button,
    register_only_instance_resource,
    register_resource,
)
from .file import FileField, ImageField, ImageURLField, UnsafeImageField
from .fsm import FSMStatusField
from .json import (
    JSONField,
    JSONTableField,
    JSONTextEditorField,
    TemplatedJSONTextEditor,
)
from .list import ListField, SparklineField
from .mixins import decorator
from .number import (
    DecimalField,
    DecimalRangeField,
    FloatField,
    IntegerField,
    YearField,
    percent_decorator,
)
from .other import EmojiRatingField, RangeSelectField, StarRatingField
from .primary_key import PrimaryKeyCharField, PrimaryKeyField
from .related import ListSerializer, PrimaryKeyRelatedField
from .text import (
    CharField,
    CodeField,
    ColorPickerField,
    MarkdownTextField,
    StringRelatedField,
    TelephoneField,
    TextAreaField,
    TextField,
    URLField,
)
from .types import ReturnContentType, WBCoreType



# Generated by Django 4.2.9 on 2024-02-27 13:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("documents", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="documentmodelrelationship",
            name="primary",
            field=models.BooleanField(default=False, verbose_name="Primary"),
        ),
        migrations.AlterUniqueTogether(
            name="documentmodelrelationship",
            unique_together={("document", "content_type", "object_id", "primary")},
        ),
    ]



from django.utils.translation import gettext_lazy as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import IntegerPreference

authentication = Section("authentication")


@global_preferences_registry.register
class HoursBeforeDeletingUnregisteredAccountPreference(IntegerPreference):
    section = authentication
    name = "hours_before_deleting_unregistered_account"
    default = 48

    verbose_name = _("Immunity Period")
    help_text = _("Set the maximum number of hours a user can remain without registration")


@global_preferences_registry.register
class DefaultTokenValidityInHoursPreference(IntegerPreference):
    section = authentication
    name = "default_token_validity_in_hours"
    default = 24

    verbose_name = _("Default Token Validity (in Hours)")
    help_text = _("If not specified, a newly created token will be valid for that amount of hours")



from django.db.models.signals import ModelSignal

get_dependant_dynamic_fields_instances = (
    ModelSignal()
)  # Signal use to gather the dependant fields to be computed before the sender (use for the dynamic field framework). Experimental state
pre_collection = (
    ModelSignal()
)  # pre_delete signal is sent after collection. Therefore, protect field will trigger an IntergrityError. This signal can be used to circuvent that.



from unittest.mock import patch

import pytest
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.workflow.models import Condition


@pytest.mark.django_db
class TestCondition:
    def test_errors_satisfied_not_called(self, condition_factory):
        condition = condition_factory()
        with pytest.raises(ValueError):
            condition.errors

    def test_errors(self, condition_factory):
        condition = condition_factory()
        with patch.object(condition, "_errors", create=True):
            assert condition.errors

    def test_satisfied_attribute_error(self, condition_factory):
        instance = PersonFactory()
        condition = condition_factory(attribute_name="Test")
        assert not condition.satisfied(instance)
        assert len(condition.errors) == 1

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_value_error(self, mock_cast, condition_factory):
        instance = PersonFactory()
        condition = condition_factory()
        mock_cast.side_effect = ValueError
        assert not condition.satisfied(instance)
        assert len(condition.errors) == 1

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_eq(self, mock_cast, condition_factory):
        instance = PersonFactory()
        condition1 = condition_factory(operator=Condition.Operator.EQ, negate_operator=False)
        condition2 = condition_factory(operator=Condition.Operator.EQ, negate_operator=True)
        mock_cast.return_value = instance.first_name
        assert condition1.satisfied(instance)
        assert not condition1.errors
        assert not condition2.satisfied(instance)
        assert not condition2.errors

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_gt(self, mock_cast, condition_factory):
        instance = PersonFactory(personality_profile_red=5)
        condition1 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.GT, negate_operator=False
        )
        condition2 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.GT, negate_operator=True
        )
        mock_cast.return_value = 3
        assert condition1.satisfied(instance)
        assert not condition1.errors
        assert not condition2.satisfied(instance)
        assert not condition2.errors

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_gte(self, mock_cast, condition_factory):
        instance = PersonFactory(personality_profile_red=5)
        condition1 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.GTE, negate_operator=False
        )
        condition2 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.GTE, negate_operator=True
        )
        mock_cast.return_value = 5
        assert condition1.satisfied(instance)
        assert not condition1.errors
        assert not condition2.satisfied(instance)
        assert not condition2.errors

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_lt(self, mock_cast, condition_factory):
        instance = PersonFactory(personality_profile_red=2)
        condition1 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.LT, negate_operator=False
        )
        condition2 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.LT, negate_operator=True
        )
        mock_cast.return_value = 5
        assert condition1.satisfied(instance)
        assert not condition1.errors
        assert not condition2.satisfied(instance)
        assert not condition2.errors

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_from_target_object")
    def test_satisfied_lte(self, mock_cast, condition_factory):
        instance = PersonFactory(personality_profile_red=2)
        condition1 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.LTE, negate_operator=False
        )
        condition2 = condition_factory(
            attribute_name="personality_profile_red", operator=Condition.Operator.LTE, negate_operator=True
        )
        mock_cast.return_value = 2
        assert condition1.satisfied(instance)
        assert not condition1.errors
        assert not condition2.satisfied(instance)
        assert not condition2.errors



import json

import tablib
from wbcore.pandas import fields as pf
from wbcore.pandas.views import PandasAPIViewSet
from wbcore.viewsets import ModelViewSet, ReadOnlyModelViewSet, RepresentationViewSet

from .configs.endpoints import (
    ImportSourceModelViewSetEndpointConfig,
    SourceModelViewSetEndpointConfig,
)
from .models import DataBackend, ImportSource, ParserHandler, Source
from .resources import ViewResource
from .serializers import (
    DataBackendRepresentationSerializer,
    ImportSourceModelSerializer,
    ImportSourceRepresentationSerializer,
    ParserHandlerModelSerializer,
    ParserHandlerRepresentationSerializer,
    SourceModelSerializer,
    SourceRepresentationSerializer,
)
from .viewset_mixins import ImportExportDRFMixin


class ImportSourceRepresentationViewSet(RepresentationViewSet):
    ordering = ["id"]
    queryset = ImportSource.objects.all()
    serializer_class = ImportSourceRepresentationSerializer


class SourceRepresentationViewSet(RepresentationViewSet):
    ordering = ["id"]
    search_fields = ["id", "title"]
    filterset_fields = {"title": ["exact", "icontains"], "id": ["in", "exact"]}
    queryset = Source.objects.all()
    serializer_class = SourceRepresentationSerializer


class ParserHandlerRepresentationViewSet(RepresentationViewSet):
    ordering = ["parser", "handler", "id"]
    search_fields = ["parser", "handler"]
    filterset_fields = {
        "parser": ["exact", "icontains"],
        "handler": ["exact", "icontains"],
    }
    queryset = ParserHandler.objects.all()
    serializer_class = ParserHandlerRepresentationSerializer


class DataBackendRepresentationViewSet(RepresentationViewSet):
    ordering = ["title", "id"]
    search_fields = ["title"]
    filterset_fields = {
        "title": ["exact", "icontains"],
        "backend_class_path": ["exact", "icontains"],
    }
    queryset = DataBackend.objects.all()
    serializer_class = DataBackendRepresentationSerializer


class ParserHandlerModelViewSet(ModelViewSet):
    ordering = ["parser", "handler", "id"]
    search_fields = ["parser", "handler"]
    filterset_fields = {
        "parser": ["exact", "icontains"],
        "handler": ["exact", "icontains"],
    }
    queryset = ParserHandler.objects.all()
    serializer_class = ParserHandlerModelSerializer


class ImportSourceModelViewSet(ModelViewSet):
    ordering = ["id"]
    search_fields = ["parser_handler__parser", "parser_handler__handler"]
    filterset_fields = {
        "parser_handler": ["exact"],
        "status": ["exact"],
        "source": ["exact"],
        "creator": ["exact"],
        "created": ["lte", "gte"],
        "last_updated": ["lte", "gte"],
    }
    queryset = ImportSource.objects.all()
    serializer_class = ImportSourceModelSerializer
    endpoint_config_class = ImportSourceModelViewSetEndpointConfig


class SourceModelViewSet(ReadOnlyModelViewSet):
    ordering = ["id"]
    search_fields = ["parser_handler__parser", "parser_handler__handler", "title", "uuid", "data_backend__title"]
    filterset_fields = {
        "title": ["icontains"],
        "uuid": ["exact"],
        "parser_handler": ["exact"],
        "is_active": ["exact"],
        "data_backend": ["exact"],
    }
    queryset = Source.objects.all()
    serializer_class = SourceModelSerializer
    endpoint_config_class = SourceModelViewSetEndpointConfig
    lookup_field = "uuid"


class ExportPandasAPIViewSet(ImportExportDRFMixin, PandasAPIViewSet):
    IMPORT_ALLOWED: bool = False

    def _get_validated_headers(self, request, df) -> list[str]:
        """
        This method validates the available columns and their ability to serialize
        """
        validated_headers = []
        pandas_fields = {f.key: f for f in self.get_pandas_fields(request).fields}
        for col in df.columns:
            if (pdf := pandas_fields.get(col, None)) and not isinstance(
                pdf, pf.SparklineField
            ):  # exclude sparkeline from exported data
                validated_headers.append(pdf.key)
        return validated_headers

    def _get_data_for_export(self, request, queryset, *args, **kwargs) -> tablib.Dataset:
        df = self._get_dataframe(export=True, **kwargs)
        validated_headers = self._get_validated_headers(request, df)
        df = df[
            df.columns.intersection(validated_headers)
        ]  # ensure df contains only the validated header, in that specific order

        columns_map = ViewResource.get_columns_map(self)
        dataset = tablib.Dataset(headers=list(columns_map.values()))
        df = json.loads(
            df.to_json(orient="records", date_format="iso")
        )  # we do that do ensure that dataframe values have be converted to json friendly data
        for data in df:
            row = []
            for key in columns_map.keys():  # Ensure data is appended to the row in the header order
                row.append(data.get(key, None))
            dataset.append(row)
        return dataset



from typing import Iterator

from wbcore.contrib.dataloader.dataloaders import DataloaderProxy

from .protocols import EntityTestDataProtocol


class EntityTestDataloaderProxy(DataloaderProxy[EntityTestDataProtocol]):
    def data(self, n: int) -> Iterator[int]:
        for dl in self.iterate_dataloaders("data"):
            yield from dl.data(n)



from contextlib import suppress

from django.db.utils import ProgrammingError
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.directory.models import Company


def get_main_company() -> Company | None:
    with suppress(RuntimeError, ProgrammingError, Company.DoesNotExist):
        return Company.objects.get(id=global_preferences_registry.manager()["directory__main_company"])



from dataclasses import dataclass
from typing import Iterable

from wbcore.utils.importlib import parse_signal_received_for_module

from .enums import NavigationType
from .pages import Page, SerializedPage
from .signals import add_display_pages

type SerializedDisplay = dict[str, NavigationType | SerializedPage]


@dataclass
class Display:
    """A display contains multiple pages and a navigation type. When only 1 page is specified, the navigation type is irrelevant"""

    pages: Iterable[Page]
    navigation_type: NavigationType = NavigationType.TAB

    def serialize(self, view_config=None, view=None, request=None, key_prefix=None) -> SerializedDisplay:
        """Serializes a `Display`

        Returns:
            A dictionary with all pages and the navigation type

        """
        serialized_pages = [
            page.serialize(key_prefix=key_prefix, view_config=view_config, view=view, request=request)
            for page in self.pages
        ]
        if view_config:
            for prefix, remote_pages in parse_signal_received_for_module(
                add_display_pages.send(view_config.__class__, request=request, view=view)
            ):
                for remote_page in remote_pages:
                    serialized_pages.append(remote_page.serialize(prefix, view_config, view, request))
        return {"navigationType": self.navigation_type.value, "pages": serialized_pages}



from random import choice

import factory
from django.db.models import signals
from faker import Faker
from wbcore.contrib.authentication.factories import GroupFactory
from wbcore.contrib.workflow.models import (
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    StartStep,
    Step,
    UserStep,
)
from wbcore.contrib.workflow.sites import workflow_site

fake = Faker()


def _generate_random_printout() -> str:
    return f'def run(process_step):\r\n   print("{fake.sentence(nb_words=5)}")'


@factory.django.mute_signals(signals.post_save)
class StepFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=128)
    workflow = factory.SubFactory("wbcore.contrib.workflow.factories.WorkflowFactory")
    code = factory.Faker("pyint", min_value=0, max_value=9999)
    status = factory.Faker("text", max_nb_chars=64)

    class Meta:
        model = Step
        django_get_or_create = ["code", "workflow"]


class RandomChildStepFactory(StepFactory):
    @classmethod
    def create(cls, **kwargs):
        child_factories = [
            StartStepFactory,
            UserStepFactory,
            DecisionStepFactory,
            SplitStepFactory,
            JoinStepFactory,
            ScriptStepFactory,
            EmailStepFactory,
            FinishStepFactory,
        ]
        if exclude_factories := kwargs.get("exclude_factories"):
            child_factory = choice(list(set(child_factories) - set(exclude_factories)))
            kwargs.pop("exclude_factories")
        else:
            child_factory = choice(child_factories)

        # Create an instance using the selected child factory
        return child_factory.create(**kwargs)

    class Meta:
        model = Step
        django_get_or_create = ["code"]


class UserStepFactory(StepFactory):
    assignee_method = factory.Iterator([i[0] for i in workflow_site.assignees_choices])
    notify_user = factory.Faker("pybool")
    group = factory.SubFactory(GroupFactory)
    step_type = Step.StepType.USERSTEP
    display = factory.SubFactory("wbcore.contrib.workflow.factories.DisplayFactory")

    class Meta:
        model = UserStep


class DecisionStepFactory(StepFactory):
    step_type = Step.StepType.DECISIONSTEP

    class Meta:
        model = DecisionStep


class SplitStepFactory(StepFactory):
    step_type = Step.StepType.SPLITSTEP

    class Meta:
        model = SplitStep


class StartStepFactory(StepFactory):
    step_type = Step.StepType.STARTSTEP

    class Meta:
        model = StartStep


class JoinStepFactory(StepFactory):
    step_type = Step.StepType.JOINSTEP
    wait_for_all = factory.Faker("pybool")

    class Meta:
        model = JoinStep


class ScriptStepFactory(StepFactory):
    step_type = Step.StepType.SCRIPTSTEP
    script = factory.LazyAttribute(lambda o: _generate_random_printout())

    class Meta:
        model = ScriptStep


class EmailStepFactory(StepFactory):
    step_type = Step.StepType.EMAILSTEP
    template = factory.django.FileField(filename="testfile.dat", data=b"testdata")
    subject = factory.Faker("text", max_nb_chars=128)

    @factory.post_generation
    def to(self, create, extracted, **kwargs):
        if not create:
            return

        elif extracted:
            for recipient in extracted:
                self.to.add(recipient)

    @factory.post_generation
    def cc(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for recipient in extracted:
                self.cc.add(recipient)

    @factory.post_generation
    def bcc(self, create, extracted, **kwargs):
        if not create:
            return

        elif extracted:
            for recipient in extracted:
                self.bcc.add(recipient)

    class Meta:
        model = EmailStep


class FinishStepFactory(StepFactory):
    step_type = Step.StepType.FINISHSTEP
    write_preserved_instance = factory.SelfAttribute("workflow.preserve_instance")

    class Meta:
        model = FinishStep



from django.apps import AppConfig
from django.db.models.signals import post_migrate

from .management import create_wbcore_permissions


class WBCoreConfig(AppConfig):
    name = "wbcore"

    def ready(self):
        """
        We load received after all apps are ready in order to be able to discover all subclasses
        """
        post_migrate.connect(
            create_wbcore_permissions,
            dispatch_uid="wbcore.management.create_wbcore_permissions",
        )



# Generated by Django 4.1.10 on 2023-08-30 11:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("workflow", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="step",
            name="step_type",
            field=models.CharField(
                choices=[
                    ("StartStep", "Start Step"),
                    ("UserStep", "User Step"),
                    ("DecisionStep", "Decision Step"),
                    ("SplitStep", "Split Step"),
                    ("JoinStep", "Join Step"),
                    ("EmailStep", "Email Step"),
                    ("ScriptStep", "Script Step"),
                    ("FinishStep", "Finish Step"),
                ],
                default="",
                max_length=64,
                verbose_name="Step Type",
            ),
        ),
    ]



from django.utils.translation import gettext as _
from wbcore.contrib.example_app.models import Team
from wbcore.metadata.configs.titles import TitleViewConfig


class SportPersonTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Persons")

    def get_create_title(self) -> str:
        return _("Create Person")

    def get_instance_title(self) -> str:
        return _("Person")


class PlayerTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Players")

    def get_create_title(self) -> str:
        return _("Create Player")

    def get_instance_title(self) -> str:
        return _("Player")


class PlayerTeamTitleConfig(PlayerTitleConfig):
    def get_create_title(self) -> str:
        if team_id := self.view.kwargs.get("team_id"):
            try:
                return _("Create Player For {}").format(Team.objects.get(id=team_id).name)
            except Team.DoesNotExist:
                pass
        return super().get_create_title()



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.workflow.models import ProcessStep, Step
from wbcore.contrib.workflow.serializers import (
    AssignedProcessStepSerializer,
    ProcessStepModelSerializer,
)
from wbcore.test.utils import get_or_create_superuser


@pytest.mark.django_db
class TestProcessStep:
    api_factory = APIRequestFactory()

    def test_next_process_step_buttons_superuser(self, transition_factory, process_step_factory, user_step_factory):
        step = user_step_factory()
        transition1 = transition_factory(
            from_step=step,
        )
        transition2 = transition_factory(
            from_step=step,
        )
        transition_factory(to_step=step)
        transition_factory()
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        transition_labels = {x["label"] for x in serializer.data["_buttons"]}
        assert transition_labels == {transition1.name, transition2.name}

    def test_next_process_step_buttons_assignee(self, transition_factory, process_step_factory, user_step_factory):
        step = user_step_factory()
        transition = transition_factory(
            from_step=step,
        )
        user = UserFactory()
        process_step = process_step_factory(assignee=user, step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = user
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"][0]["label"] == transition.name

    def test_next_process_step_buttons_group(self, transition_factory, process_step_factory, user_step_factory):
        step = user_step_factory(assignee_method=None)
        transition = transition_factory(
            from_step=step,
        )
        group = GroupFactory()
        user = UserFactory(groups=[group])
        process_step = process_step_factory(group=group, step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = user
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"][0]["label"] == transition.name

    @pytest.mark.parametrize(
        "state",
        [
            ProcessStep.StepState.CANCELED,
            ProcessStep.StepState.FAILED,
            ProcessStep.StepState.FINISHED,
            ProcessStep.StepState.WAITING,
        ],
    )
    def test_next_process_step_buttons_not_active(
        self, state, transition_factory, process_step_factory, user_step_factory
    ):
        step = user_step_factory()
        transition_factory(
            from_step=step,
        )
        process_step = process_step_factory(step=step, state=state)
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"] == []

    @pytest.mark.parametrize(
        "step_type",
        [
            Step.StepType.DECISIONSTEP,
            Step.StepType.EMAILSTEP,
            Step.StepType.FINISHSTEP,
            Step.StepType.JOINSTEP,
            Step.StepType.SCRIPTSTEP,
            Step.StepType.SPLITSTEP,
        ],
    )
    def test_next_process_step_buttons_not_user_step(
        self, step_type, transition_factory, process_step_factory, step_factory
    ):
        step = step_factory(step_type=step_type)
        transition_factory(
            from_step=step,
        )
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"] == []

    def test_next_process_step_buttons_not_assigned(self, transition_factory, process_step_factory, user_step_factory):
        step = user_step_factory()
        transition_factory(
            from_step=step,
        )
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = UserFactory()
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"] == []

    def test_next_process_step_buttons_assigned_group_with_method(
        self, transition_factory, process_step_factory, user_step_factory
    ):
        step = user_step_factory()
        transition_factory(
            from_step=step,
        )
        group = GroupFactory()
        user = UserFactory(groups=[group])
        process_step = process_step_factory(group=group, step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = user
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"] == []

    def test_next_process_step_buttons_no_transitions(
        self, transition_factory, process_step_factory, user_step_factory
    ):
        step = user_step_factory()
        transition_factory()
        transition_factory(to_step=step)
        transition_factory()
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = ProcessStepModelSerializer(process_step, context={"request": request})
        assert serializer.data
        assert serializer.data["_buttons"] == []


@pytest.mark.django_db
class TestAssignedProcessStep:
    api_factory = APIRequestFactory()

    def test_get_instance_endpoint_no_request(self, process_step_factory):
        process_step = process_step_factory()
        serializer = AssignedProcessStepSerializer(process_step)
        assert serializer.data
        assert serializer.data["instance_endpoint"] == ""

    def test_get_instance_endpoint_no_attached_instance(self, process_step_factory):
        process_step = process_step_factory(process__instance=None, process__instance_id=None)
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = AssignedProcessStepSerializer(process_step, context={"request": request})
        assert serializer.data
        assert "processstep" in serializer.data["instance_endpoint"]

    def test_get_instance_endpoint_attached_instance(self, process_step_factory):
        process_step = process_step_factory()
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        serializer = AssignedProcessStepSerializer(process_step, context={"request": request})
        assert serializer.data
        assert "processstep" not in serializer.data["instance_endpoint"]
        assert "person" in serializer.data["instance_endpoint"]



from rest_framework.reverse import reverse


def resolve_profile(request):
    user = request.user
    return {
        "image": user.profile.profile_image.url if user.profile.profile_image else None,
        "endpoint": reverse("wbcore:authentication:userprofile-detail", args=[user.id], request=request),
    }



import pandas as pd
import plotly.graph_objects as go
import tablib
from django.db.models import (
    Case,
    Count,
    Exists,
    F,
    IntegerField,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    When,
)
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.example_app.filters import (
    EventFilter,
    EventMatchFilter,
    EventTypeFilter,
    EventTypeSportFilter,
    PlayerStatisticsChartFilter,
)
from wbcore.contrib.example_app.models import (
    Event,
    EventType,
    Match,
    Player,
    SportPerson,
)
from wbcore.contrib.example_app.serializers import (
    EventModelSerializer,
    EventRepresentationSerializer,
    EventTypeModelSerializer,
    EventTypeRepresentationSerializer,
    LeaguePlayerStatisticsModelSerializer,
    LeagueTeamStatisticsModelSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import (
    EventDisplayConfig,
    EventMatchDisplayConfig,
    EventTypeDisplayConfig,
    EventTypeSportDisplayConfig,
    LeaguePlayerStatisticsDisplayConfig,
    LeagueTeamStatisticsDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import (
    EventEndpointConfig,
    EventMatchEndpointConfig,
    EventTypeSportEndpointConfig,
    LeagueStatisticsEndpointConfig,
)
from wbcore.contrib.example_app.viewsets.titles import (
    EventTitleConfig,
    EventTypeSportTitleConfig,
    EventTypeTitleConfig,
    PlayerStatisticsChartTitleConfig,
)


class EventModelViewSet(viewsets.ModelViewSet):
    display_config_class = EventDisplayConfig
    queryset = Event.objects.all()
    search_fields = ("person__computed_str", "event_type__name")
    serializer_class = EventModelSerializer
    title_config_class = EventTitleConfig
    ordering_fields = ("match__home", "minute", "person__computed_str", "event_type__name")
    ordering = ("match__id", "minute", "person__computed_str", "event_type__name")
    filterset_class = EventFilter
    endpoint_config_class = EventEndpointConfig

    @action(detail=False, methods=["PATCH"])
    def matchevent(self, request, pk=None):
        person_id = request.POST.get("person")
        minute = request.POST.get("minute")
        event_type_id = request.GET.get("event_type")
        match_id = request.GET.get("match")
        data = {"person": person_id, "minute": minute, "event_type": event_type_id, "match": match_id}
        serializer = self.serializer_class(data=data, context={"request": request})
        if serializer.is_valid():
            person = get_object_or_404(SportPerson, pk=person_id)
            event_type = get_object_or_404(EventType, pk=event_type_id)
            match = get_object_or_404(Match, pk=match_id)
            event = Event(person=person, minute=minute, event_type=event_type, match=match)
            event.save()
            return Response("Event Created", status=status.HTTP_200_OK)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def get_queryset(self) -> QuerySet[Event]:
        return super().get_queryset().select_related("person", "event_type", "match")


class PlayerStatisticsChartModelViewSet(viewsets.ChartViewSet):
    EXPORT_ALLOWED = True
    queryset = Event.objects.all()
    title_config_class = PlayerStatisticsChartTitleConfig
    filterset_class = PlayerStatisticsChartFilter

    def get_queryset(self):
        return super().get_queryset().filter(person=self.kwargs["player_id"])

    def _get_data_for_export(self, queryset, **kwargs) -> tablib.Dataset:
        dataset = tablib.Dataset(headers=["League", "Event"])
        for event in queryset.select_related("event_type", "match", "match__league"):
            dataset.append([event.match.league.name, event.event_type.name])
        return dataset

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.exists():
            df = pd.DataFrame(queryset.values("match__league__name", "event_type__name"))
            df.fillna("None", inplace=True)
            pivot_df = df.pivot_table(
                index="match__league__name",
                columns="event_type__name",
                aggfunc="size",
                fill_value=0,
            )

            for event_type in pivot_df.columns:
                fig.add_trace(
                    go.Histogram(
                        histfunc="sum",
                        y=pivot_df[event_type],
                        x=pivot_df.index,
                        name=f"{event_type}s",
                    )
                )

            fig.update_layout(
                xaxis=dict(title=_("League")),
                yaxis=dict(title=_("Amount"), type="linear"),
                autosize=True,
                bargap=0.8,
                bargroupgap=0.7,
            )

        return fig


class LeaguePlayerStatisticsModelViewSet(viewsets.ModelViewSet):
    endpoint_config_class = LeagueStatisticsEndpointConfig
    serializer_class = LeaguePlayerStatisticsModelSerializer
    display_config_class = LeaguePlayerStatisticsDisplayConfig
    search_fields = ("person_name",)
    ordering_fields = ("person_name", "count")
    ordering = ("-count", "person_name", "person_id")

    def get_queryset(self) -> QuerySet[Event]:
        return (
            Event.objects.filter(
                match__league=self.kwargs["league_id"],
                event_type=self.kwargs["event_type_id"],
            )
            .values("person")
            .annotate(
                count=Count("person"),
                person_name=F("person__computed_str"),
                person_id=F("person_id"),
                id=F("person_id"),
            )
        )


class LeagueTeamStatisticsModelViewSet(viewsets.ModelViewSet):
    endpoint_config_class = LeagueStatisticsEndpointConfig
    serializer_class = LeagueTeamStatisticsModelSerializer
    display_config_class = LeagueTeamStatisticsDisplayConfig
    search_fields = ("team_name",)
    ordering_fields = ("team_name", "count")
    ordering = ("-count", "team_name", "team_id")

    def get_queryset(self) -> QuerySet[Event]:
        return (
            Event.objects.filter(
                Q(match__league=self.kwargs["league_id"])
                & Q(event_type=self.kwargs["event_type_id"])
                & Q(
                    Q(person__coached_team__isnull=False)
                    | Q(person__in=Player.objects.filter(current_team__isnull=False).values("id"))
                )
            )
            .values("person")
            .annotate(
                is_player=Exists(Player.objects.filter(id=OuterRef("person_id"), current_team__isnull=False)),
                is_coach=Q(person__coached_team__isnull=False),
                team_id=Case(
                    When(
                        is_player=True,
                        then=Subquery(Player.objects.filter(id=OuterRef("person_id")).values("current_team__id")[:1]),
                    ),
                    When(
                        is_coach=True,
                        then=F("person__coached_team__id"),
                    ),
                    output_field=IntegerField(),
                ),
            )
            .values("team_id")
            .annotate(
                count=Count("person"),
                id=F("team_id"),
                team_name=Case(
                    When(
                        is_player=True,
                        then=Subquery(
                            Player.objects.filter(id=OuterRef("person_id")).values("current_team__name")[:1]
                        ),
                    ),
                    When(
                        is_coach=True,
                        then=F("person__coached_team__name"),
                    ),
                ),
            )
        )


class EventMatchModelViewSet(EventModelViewSet):
    display_config_class = EventMatchDisplayConfig
    ordering_fields = ("minute", "person__computed_str", "event_type__name")
    ordering = ("minute", "person__computed_str", "event_type__name")
    filterset_class = EventMatchFilter
    endpoint_config_class = EventMatchEndpointConfig

    def get_queryset(self) -> QuerySet[Event]:
        queryset = super().get_queryset()
        if match_id := self.kwargs.get("match_id"):
            queryset = queryset.filter(match=match_id)
        return queryset


class EventRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Event.objects.all()
    serializer_class = EventRepresentationSerializer
    search_fields = ("event_type__name", "minute", "person__computed_str")


class EventTypeModelViewSet(viewsets.ModelViewSet):
    display_config_class = EventTypeDisplayConfig
    queryset = EventType.objects.all()
    search_fields = ("name", "sport")
    serializer_class = EventTypeModelSerializer
    title_config_class = EventTypeTitleConfig
    ordering_fields = (
        "name",
        "points",
        "sport__name",
        "icon",
        "color",
    )
    ordering = ("sport", "name")
    filterset_class = EventTypeFilter

    def get_queryset(self) -> QuerySet[EventType]:
        return super().get_queryset().select_related("sport")


class EventTypeSportModelViewSet(EventTypeModelViewSet):
    display_config_class = EventTypeSportDisplayConfig
    ordering_fields = (
        "name",
        "points",
        "icon",
        "color",
    )
    ordering = ("name",)
    filterset_class = EventTypeSportFilter
    endpoint_config_class = EventTypeSportEndpointConfig
    title_config_class = EventTypeSportTitleConfig

    def get_queryset(self) -> QuerySet[EventType]:
        queryset = super().get_queryset()
        if sport_id := self.kwargs.get("sport_id"):
            queryset = queryset.filter(sport=sport_id)
        return queryset


class EventTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = EventType.objects.all()
    serializer_class = EventTypeRepresentationSerializer
    search_fields = ("name",)



from typing import Any, Callable

from langchain_core.language_models import BaseChatModel
from langchain_core.messages.base import BaseMessage
from langchain_core.output_parsers import StrOutputParser
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

type Prompt = str | list[BaseMessage] | BaseMessage


def construct_prompt(instance: Any, prompt: Prompt | Callable[[Any], Prompt]) -> Prompt:
    if isinstance(prompt, str):
        return prompt.format(instance)

    elif callable(prompt):
        return prompt(instance)

    return prompt


def run_llm(
    prompt: Prompt | Callable[[Any], Prompt],
    instance: Any | None = None,
    output_model: "type[BaseModel] | None" = None,
    chat_model: type[BaseChatModel] = ChatOpenAI,
    chat_model_name: str = "gpt-4o-mini",
    max_tokens: int = 16000,
    **kwargs,
) -> str | dict[str, Any]:
    model = chat_model(model=chat_model_name, max_tokens=max_tokens, **kwargs)  # type: ignore
    prompt = construct_prompt(instance, prompt)
    if output_model is None:
        result = model.invoke(prompt, **kwargs)  # type: ignore
        parser = StrOutputParser()
        return parser.invoke(result, **kwargs)

    structured_model = model.with_structured_output(output_model)
    result = structured_model.invoke(prompt)  # type: ignore

    output = dict()
    for field, value in result:
        output[field] = value

    return output



from wbcore import viewsets
from wbcore.contrib.workflow.filters import (
    DecisionStepFilter,
    EmailStepFilter,
    FinishStepFilter,
    JoinStepFilter,
    ScriptStepFilter,
    SplitStepFilter,
    StartStepFilter,
    StepFilter,
    StepWorkflowFilter,
    UserStepFilter,
)
from wbcore.contrib.workflow.models import (
    DecisionStep,
    Display,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    StartStep,
    Step,
    UserStep,
)
from wbcore.contrib.workflow.serializers import (
    DecisionStepModelSerializer,
    DisplayModelSerializer,
    EmailStepModelSerializer,
    FinishStepModelSerializer,
    JoinStepModelSerializer,
    ScriptStepModelSerializer,
    SplitStepModelSerializer,
    StartStepModelSerializer,
    StepModelSerializer,
    UserStepModelSerializer,
)
from wbcore.contrib.workflow.viewsets.buttons import StepButtonConfig
from wbcore.contrib.workflow.viewsets.display import (
    BaseStepDisplayConfig,
    EmailStepDisplayConfig,
    FinishDisplayConfig,
    JoinStepDisplayConfig,
    ScriptStepDisplayConfig,
    StartStepDisplay,
    StepDisplayConfig,
    UserStepDisplayConfig,
)
from wbcore.contrib.workflow.viewsets.endpoints import (
    DisplayEndpointConfig,
    StepEndpointConfig,
)
from wbcore.contrib.workflow.viewsets.titles import (
    DecisionStepTitleConfig,
    EmailStepTitleConfig,
    FinishStepTitleConfig,
    JoinStepTitleConfig,
    ScriptStepTitleConfig,
    SplitStepTitleConfig,
    StartStepTitleConfig,
    StepTitleConfig,
    UserStepTitleConfig,
)
from wbcore.metadata.configs.buttons import ButtonViewConfig


class StepModelViewSet(viewsets.ModelViewSet):
    queryset = Step.objects.all()
    serializer_class = StepModelSerializer
    display_config_class = StepDisplayConfig
    search_fields = ("name",)
    ordering = ("workflow__name", "name")
    endpoint_config_class = StepEndpointConfig
    title_config_class = StepTitleConfig
    button_config_class = StepButtonConfig
    ordering_fields = (
        "name",
        "workflow__name",
        "code",
        "status",
        "step_type",
    )

    def get_queryset(self):
        # TODO: OPTIMIZE PERFORMANCE, 390 duplicate queries
        queryset = super().get_queryset().select_related("workflow")
        if workflow_id := self.kwargs.get("workflow_id"):
            return queryset.filter(workflow=workflow_id)
        return queryset

    def get_filterset_class(self, request):
        if self.kwargs.get("workflow_id"):
            return StepWorkflowFilter
        return StepFilter


class UserStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/userstep.md"
    queryset = UserStep.objects.all()
    serializer_class = UserStepModelSerializer
    display_config_class = UserStepDisplayConfig
    title_config_class = UserStepTitleConfig
    search_fields = ("name", "assignee", "group")
    button_config_class = ButtonViewConfig
    ordering = ("workflow__name", "assignee__email", "name")
    ordering_fields = StepModelViewSet.ordering_fields + (
        "assignee__computed_str",
        "group__name",
        "notify_user",
        "permission__name",
        "assignee_method",
    )

    def get_queryset(self):
        return super().get_queryset().select_related("assignee")

    def get_filterset_class(self, request):
        return UserStepFilter


class DecisionStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/decisionstep.md"
    queryset = DecisionStep.objects.all()
    title_config_class = DecisionStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = BaseStepDisplayConfig
    serializer_class = DecisionStepModelSerializer

    def get_filterset_class(self, request):
        return DecisionStepFilter


class StartStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/startstep.md"
    queryset = StartStep.objects.all()
    title_config_class = StartStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = StartStepDisplay
    serializer_class = StartStepModelSerializer

    def get_filterset_class(self, request):
        return StartStepFilter


class SplitStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/splitstep.md"
    queryset = SplitStep.objects.all()
    title_config_class = SplitStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = BaseStepDisplayConfig
    serializer_class = SplitStepModelSerializer

    def get_filterset_class(self, request):
        return SplitStepFilter


class JoinStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/joinstep.md"
    queryset = JoinStep.objects.all()
    title_config_class = JoinStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = JoinStepDisplayConfig
    serializer_class = JoinStepModelSerializer
    ordering = ("workflow__name", "wait_for_all", "name")
    ordering_fields = StepModelViewSet.ordering_fields + ("wait_for_all",)

    def get_filterset_class(self, request):
        return JoinStepFilter


class ScriptStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/scriptstep.md"
    queryset = ScriptStep.objects.all()
    title_config_class = ScriptStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = ScriptStepDisplayConfig
    serializer_class = ScriptStepModelSerializer
    ordering_fields = StepModelViewSet.ordering_fields + ("script",)

    def get_filterset_class(self, request):
        return ScriptStepFilter


class EmailStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/emailstep.md"
    queryset = EmailStep.objects.all()
    title_config_class = EmailStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = EmailStepDisplayConfig
    serializer_class = EmailStepModelSerializer
    ordering_fields = StepModelViewSet.ordering_fields + (
        "subject",
        "to__adress",
        "cc__adress",
        "bcc__address",
        "template__name",
    )
    search_fields = ("name", "subject", "to__address")

    def get_filterset_class(self, request):
        return EmailStepFilter

    def get_queryset(self):
        return super().get_queryset().prefetch_related("to", "cc", "bcc")


class FinishStepModelViewSet(StepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/finishstep.md"
    queryset = FinishStep.objects.all()
    title_config_class = FinishStepTitleConfig
    button_config_class = ButtonViewConfig
    display_config_class = FinishDisplayConfig
    serializer_class = FinishStepModelSerializer
    ordering_fields = StepModelViewSet.ordering_fields + ("write_preserved_instance",)

    def get_filterset_class(self, request):
        return FinishStepFilter


class DisplayModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/display.md"
    queryset = Display.objects.all()
    serializer_class = DisplayModelSerializer
    endpoint_config_class = DisplayEndpointConfig
    search_fields = ordering = ("name",)
    ordering_fields = (
        "name",
        "grid_template_areas",
    )



import pytest
from bs4 import BeautifulSoup
from django.core import mail
from django.test.utils import override_settings
from wbcore.contrib.notifications import tasks


@pytest.fixture
def patched_send_notification_email(mocker):
    return mocker.patch("wbcore.contrib.notifications.tasks.send_notification_email")


@pytest.fixture
def patched_import_string(mocker):
    return mocker.patch("wbcore.contrib.notifications.tasks.import_string")


@pytest.fixture
def mock_backend():
    from wbcore.contrib.notifications.backends.abstract_backend import (
        AbstractNotificationBackend,
    )

    class MockedBackend(AbstractNotificationBackend):
        @classmethod
        def send_notification(cls, notification):
            pass

        @classmethod
        def get_configuration(cls) -> dict:
            return {}

    return MockedBackend


@pytest.mark.django_db
def test_send_notification_task(
    notification, mock_backend, mocker, patched_send_notification_email, patched_import_string
):
    notification.user.wbnotification_user_settings.update(enable_email=True)
    spy = mocker.spy(mock_backend, "send_notification")
    patched_import_string.return_value = mock_backend

    tasks.send_notification_task(notification.pk)

    patched_send_notification_email.assert_called_once_with(notification)
    patched_import_string.assert_called_once()
    spy.assert_called_once_with(notification)


@pytest.mark.django_db
def test_send_notification_task_without_mail(
    notification, mock_backend, patched_send_notification_email, patched_import_string
):
    notification.user.wbnotification_user_settings.update(enable_email=False)
    patched_import_string.return_value = mock_backend

    tasks.send_notification_task(notification.pk)

    assert not patched_send_notification_email.called


@pytest.mark.django_db
@override_settings(EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
@pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
def test_send_notification_email(notification):
    assert len(mail.outbox) == 0
    tasks.send_notification_email(notification)
    assert len(mail.outbox) == 1
    soup = BeautifulSoup(mail.outbox[0].alternatives[0][0], "html.parser")
    assert soup.find("a", href=notification.get_full_endpoint(as_shareable_internal_link=True))



import functools
from typing import TYPE_CHECKING

from django.conf import settings
from django.utils import timezone
from rest_framework import HTTP_HEADER_ENCODING
from rest_framework.authentication import TokenAuthentication as DRFTokenAuthentication
from rest_framework.request import Request
from rest_framework_simplejwt.authentication import JWTAuthentication
from rest_framework_simplejwt.settings import api_settings

from .models import Token

AUTH_HEADER_TYPES = api_settings.AUTH_HEADER_TYPES

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User
else:
    from django.contrib.auth import get_user_model

    User = get_user_model()


class TokenAuthentication(DRFTokenAuthentication):
    """
    Short lived Token token based authentication.

    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "Token ".  For example:

        Authorization: Token 401f7ac837da42b97f613d789819ff93537bee6a
    """

    model = Token

    # TODO: Find a way to improve typing in the next two methods, somehow pyright is not picking up self.model
    def authenticate_credentials(self, key) -> tuple[User, Token]:
        return super().authenticate_credentials(key)  # type: ignore

    def authenticate(self, request) -> tuple[User, Token] | None:
        authentication_res = super().authenticate(request)
        if authentication_res:
            authentication_res[1].check_validity_for_endpoint(request.resolver_match.view_name, timezone.now())  # type: ignore
        return authentication_res  # type: ignore


class QueryTokenAuthentication(TokenAuthentication):
    """
    Short lived Token token based authentication through query parameters.

    Clients should authenticate by passing the token key "token" as query parameters. For example:
        ?token=401f7ac837da42b97f613d789819ff93537bee6a
    Note, this is unsafe to use this authentication backend on unsecured connection
    """

    query_param_name = "token"

    # TODO: Find a way to improve typing in the next method, somehow pyright is not picking up self.model
    def authenticate(self, request) -> tuple[User, Token] | None:
        token = request.query_params.get(self.query_param_name)
        if not token:
            return None
        user, token = self.authenticate_credentials(token)
        token.check_validity_for_endpoint(request.resolver_match.view_name, timezone.now())
        return user, token  # type: ignore


class JWTCookieAuthentication(JWTAuthentication):
    def get_header(self, request: Request) -> bytes:
        if cookie_token := request.COOKIES.get(settings.JWT_COOKIE_KEY, None):
            return f"{AUTH_HEADER_TYPES[0]} {cookie_token}".encode(HTTP_HEADER_ENCODING)
        return super().get_header(request)


def inject_short_lived_token(view_name: str | None = None):
    """
    Decorator to wrap around additional resource function which return key value pair of resources and endpoint.
    The decorator will create for the user a short lived token only valid for the requested endpoint and inject it as query parameters

    The decorator expects a view_name (namespace:view_name) to be given otherwise, it will guess it from the resource endpoint (through the resolver)
    Args:
        view_name: Optional, view name to use instead of resolved the endpoint view name
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(serializer, instance, request, user, **kwargs):
            res = {}
            if view_name is None:
                return res
            for key, endpoint in func(serializer, instance, request, user, **kwargs).items():
                res[key] = Token.generate_shareable_link(endpoint, user, protected_view_name=view_name)
            return res

        return wrapper

    return decorator


def unauthenticated(request: "Request") -> dict[str, None]:
    return {"type": None}


def get_dev_user_from_settings(username_field_key: str) -> list[dict[str, str]]:
    if dev_user := getattr(settings, "DEV_USERS", []):
        users = list()
        for user in dev_user:
            username, password = user.split(":")
            users.append({username_field_key: username, "password": password})
        return users
    return []



from unittest.mock import patch

import pytest
from django.template import TemplateSyntaxError
from wbcore.contrib.directory.factories import EmailContactFactory


@pytest.mark.django_db
class TestEmailStep:
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    @patch("django.core.mail.EmailMultiAlternatives.__init__")
    @patch("django.core.mail.EmailMultiAlternatives.send")
    @patch("django.core.mail.EmailMultiAlternatives.attach_alternative")
    def test_run(self, mock_attach, mock_send, mock_init, mock_execute, process_step_factory, email_step_factory):
        step = email_step_factory(to=[EmailContactFactory()], cc=[EmailContactFactory(), EmailContactFactory()])
        process_step = process_step_factory(step=step)
        mock_init.return_value = None
        step.run(process_step)
        assert mock_init.call_args.args[0] == step.subject
        assert mock_init.call_args.kwargs == {
            "to": list(step.to.values_list("address", flat=True)),
            "cc": list(step.cc.values_list("address", flat=True)),
            "bcc": [],
        }
        assert mock_attach.called
        assert mock_send.called
        assert mock_execute.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    @patch("django.template.Template.render")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    @patch("django.core.mail.EmailMultiAlternatives.send")
    def test_run_failed(
        self, mock_send, mock_execute, mock_render, mock_set_failed, process_step_factory, email_step_factory
    ):
        step = email_step_factory(
            to=[EmailContactFactory()],
            cc=[EmailContactFactory(), EmailContactFactory()],
        )
        process_step = process_step_factory(step=step)
        mock_render.side_effect = TemplateSyntaxError
        step.run(process_step)
        assert not mock_send.called
        assert not mock_execute.called
        assert mock_set_failed.call_args.args[0] == process_step



from django.utils.translation import gettext_lazy as _
from wbcore import filters
from wbcore.contrib.example_app.models import (
    League,
    Match,
    Sport,
    SportPerson,
    Stadium,
    Team,
)


class MatchBaseFilter(filters.FilterSet):
    date_time__gte = filters.DateTimeFilter(
        label=_("Date Time"),
        lookup_expr="gte",
        field_name="date_time",
    )
    date_time__lte = filters.DateTimeFilter(
        label=_("Date Time"),
        lookup_expr="lte",
        field_name="date_time",
    )
    home = filters.ModelChoiceFilter(
        label=_("Home"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    away = filters.ModelChoiceFilter(
        label=_("Away"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    score_home__gte = filters.NumberFilter(label=_("Home Score"), lookup_expr="gte", field_name="score_home")
    score_home__lte = filters.NumberFilter(label=_("Home Score"), lookup_expr="lte", field_name="score_home")
    score_away__gte = filters.NumberFilter(label=_("Away Score"), lookup_expr="gte", field_name="score_away")
    score_away__lte = filters.NumberFilter(label=_("Away Score"), lookup_expr="lte", field_name="score_away")
    referee = filters.ModelMultipleChoiceFilter(
        label=_("Referees"),
        queryset=SportPerson.objects.all(),
        endpoint=SportPerson.get_representation_endpoint(),
        value_key=SportPerson.get_representation_value_key(),
        label_key=SportPerson.get_representation_label_key(),
    )
    status = filters.ChoiceFilter(
        label=_("Status"), choices=Match.MatchStatus.choices, default=Match.MatchStatus.ONGOING
    )
    sport = filters.ModelMultipleChoiceFilter(
        label=_("Sports"),
        queryset=Sport.objects.all(),
        endpoint=Sport.get_representation_endpoint(),
        value_key=Sport.get_representation_value_key(),
        label_key=Sport.get_representation_label_key(),
    )

    class Meta:
        model = Match
        fields = {}


class MatchLeagueFilter(MatchBaseFilter):
    stadium = filters.ModelMultipleChoiceFilter(
        label=_("Stadiums"),
        queryset=Stadium.objects.all(),
        endpoint=Stadium.get_representation_endpoint(),
        value_key=Stadium.get_representation_value_key(),
        label_key=Stadium.get_representation_label_key(),
    )


class MatchStadiumFilter(MatchBaseFilter):
    league = filters.ModelMultipleChoiceFilter(
        label=_("Leagues"),
        queryset=League.objects.all(),
        endpoint=League.get_representation_endpoint(),
        value_key=League.get_representation_value_key(),
        label_key=League.get_representation_label_key(),
    )


class MatchFilter(MatchStadiumFilter):
    stadium = filters.ModelMultipleChoiceFilter(
        label=_("Stadiums"),
        queryset=Stadium.objects.all(),
        endpoint=Stadium.get_representation_endpoint(),
        value_key=Stadium.get_representation_value_key(),
        label_key=Stadium.get_representation_label_key(),
    )



from django.utils.translation import gettext as _
from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.models import Group, Permission, User
from wbcore.contrib.authentication.serializers import (
    GroupRepresentationSerializer,
    PermissionRepresentationSerializer,
    UserRepresentationSerializer,
)
from wbcore.contrib.directory.models import EmailContact
from wbcore.contrib.directory.serializers import EmailContactRepresentationSerializer
from wbcore.contrib.workflow.models import (
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    StartStep,
    Step,
    UserStep,
    Workflow,
)
from wbcore.contrib.workflow.serializers.display import DisplayRepresentationSerializer
from wbcore.contrib.workflow.serializers.workflow import (
    WorkflowRepresentationSerializer,
)
from wbcore.contrib.workflow.sites import workflow_site


class StepModelSerializer(wb_serializers.ModelSerializer):
    workflow = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.DefaultFromGET("workflow"),
        queryset=Workflow.objects.all(),
        label=_("Workflow"),
    )
    _workflow = WorkflowRepresentationSerializer(source="workflow")
    _permission = PermissionRepresentationSerializer(source="permission")
    casted_endpoint = wb_serializers.SerializerMethodField()

    @wb_serializers.register_resource()
    def transitions_inline(self, instance: Step, request, user) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "transitions": reverse(
                "wbcore:workflow:transition-step-list", args=[instance.pk], request=self.context["request"]
            )
        }

    @wb_serializers.register_resource()
    def process_steps_inline(self, instance: Step, request, user) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "process_steps": reverse(
                "wbcore:workflow:processstep-step-list", args=[instance.pk], request=self.context["request"]
            )
        }

    def get_casted_endpoint(self, obj) -> str:
        if not self.context.get("request"):
            return ""
        return reverse(
            f"{obj.get_casted_step().get_endpoint_basename()}-detail", args=[obj.pk], request=self.context["request"]
        )

    def validate(self, data: dict) -> dict:
        name: str | None = data.get("name", self.instance.name if self.instance else None)
        workflow: Workflow | None = data.get("workflow", self.instance.workflow if self.instance else None)
        code: int | None = data.get("code", self.instance.code if self.instance else None)
        if "status" in data:
            status: str | None = data["status"]
        else:
            status: str | None = self.instance.status if self.instance else None

        if not workflow:
            raise ValidationError({"workflow": _("You need to set a workflow.")})

        step = Step.objects.filter(name=name, workflow=workflow)
        if self.instance:
            step = step.exclude(id=self.instance.id)
        if step.exists():
            raise ValidationError({"name": _("Name has to be unique for the specified workflow.")})

        step = Step.objects.filter(code=code, workflow=workflow)
        if self.instance:
            step = step.exclude(id=self.instance.id)
        if step.exists():
            raise ValidationError({"code": _("Code has to be unique for the specified workflow.")})

        if bool(status) ^ bool(workflow.model):
            raise ValidationError(
                {
                    "status": _(
                        "You need to specify a status for a workflow with an attached model. If no model is attached, do not specify a status."
                    )
                }
            )
        return data

    class Meta:
        model = Step
        fields = (
            "id",
            "name",
            "workflow",
            "_workflow",
            "code",
            "status",
            "step_type",
            "permission",
            "_permission",
            "_additional_resources",
            "casted_endpoint",
        )
        read_only_fields = ("step_type",)


class StepRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:step-detail")

    def get_filter_params(self, request):
        filter_params = {}
        if (view := request.parser_context["view"]) and (workflow := view.workflow):
            filter_params["workflow"] = workflow.id
        return filter_params

    class Meta:
        model = Step
        fields = (
            "id",
            "name",
            "_detail",
        )


class UserStepModelSerializer(StepModelSerializer):
    _assignee = UserRepresentationSerializer(source="assignee")
    _group = GroupRepresentationSerializer(source="group")
    _display = DisplayRepresentationSerializer(source="display")
    assignee_method = wb_serializers.ChoiceField(label=_("Assignee Method"), choices=workflow_site.assignees_choices)

    def validate(self, data: dict) -> dict:
        data = super().validate(data)
        if "assignee" in data:
            assignee: User | None = data["assignee"]
        else:
            assignee: User | None = self.instance.assignee if self.instance else None
        if "assignee_method" in data:
            assignee_method: str | None = data["assignee_method"]
        else:
            assignee_method: str | None = self.instance.assignee_method if self.instance else None
        if "permission" in data:
            permission: Permission | None = data["permission"]
        else:
            permission: Permission | None = self.instance.permission if self.instance else None
        if "group" in data:
            group: Group | None = data["group"]
        else:
            group: Group | None = self.instance.group if self.instance else None

        if assignee_method and assignee_method not in map(lambda x: x[0], workflow_site.assignees_choices):
            raise ValidationError({"assignee_method": _("Assignee method must be one of the available choices.")})

        if assignee:
            if group:
                raise ValidationError({"group": _("Please select either an assignee or a group. Do not select both.")})
            if assignee_method:
                raise ValidationError({"assignee_method": _("Please do not select both an assignee and a method.")})

        if permission and (assignee or group):
            all_assignees = []
            if assignee:
                all_assignees = [assignee]
            elif group:
                if permission in group.permissions.all():
                    return data
                all_assignees = list(group.user_set.all())
            for user in all_assignees:
                if user.is_superuser or permission in user.user_permissions.all():
                    return data
            raise ValidationError(
                {
                    "permission": _(
                        "None of the selected assignees/group members has this permission. Please assign the permission to one of the users or choose a different one."
                    )
                }
            )
        return data

    class Meta(StepModelSerializer.Meta):
        model = UserStep
        fields = StepModelSerializer.Meta.fields + (
            "assignee",
            "_assignee",
            "group",
            "_group",
            "notify_user",
            "display",
            "_display",
            "assignee_method",
            "kwargs",
        )
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class UserStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:userstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = UserStep
        fields = StepRepresentationSerializer.Meta.fields


class StartStepModelSerializer(StepModelSerializer):
    class Meta(StepModelSerializer.Meta):
        model = StartStep
        fields = StepModelSerializer.Meta.fields
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class StartStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:startstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = StartStep
        fields = StepRepresentationSerializer.Meta.fields


class DecisionStepModelSerializer(StepModelSerializer):
    class Meta(StepModelSerializer.Meta):
        model = DecisionStep
        fields = StepModelSerializer.Meta.fields
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class DecisionStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:decisionstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = DecisionStep
        fields = StepRepresentationSerializer.Meta.fields


class SplitStepModelSerializer(StepModelSerializer):
    class Meta(StepModelSerializer.Meta):
        model = SplitStep
        fields = StepModelSerializer.Meta.fields
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class SplitStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:splitstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = SplitStep
        fields = StepRepresentationSerializer.Meta.fields


class JoinStepModelSerializer(StepModelSerializer):
    class Meta(StepModelSerializer.Meta):
        model = JoinStep
        fields = StepModelSerializer.Meta.fields + ("wait_for_all",)
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class JoinStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:joinstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = JoinStep
        fields = StepRepresentationSerializer.Meta.fields


class ScriptStepModelSerializer(StepModelSerializer):
    script = wb_serializers.CodeField(label=_("Script"))

    class Meta(StepModelSerializer.Meta):
        model = ScriptStep
        fields = StepModelSerializer.Meta.fields + ("script",)
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class ScriptStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:scriptstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = ScriptStep
        fields = StepRepresentationSerializer.Meta.fields


class EmailStepModelSerializer(StepModelSerializer):
    _to = EmailContactRepresentationSerializer(source="to", many=True)
    _cc = EmailContactRepresentationSerializer(source="cc", many=True)
    _bcc = EmailContactRepresentationSerializer(source="bcc", many=True)

    def validate(self, data: dict) -> dict:
        to: list[EmailContact] = data.get("to", list(self.instance.to.all()) if self.instance else [])
        if "cc" in data:
            cc: list[EmailContact] = data["cc"]
        else:
            cc: list[EmailContact] = list(self.instance.cc.all()) if self.instance else []
        if "bcc" in data:
            bcc: list[EmailContact] = data["bcc"]
        else:
            bcc: list[EmailContact] = list(self.instance.bcc.all()) if self.instance else []

        to = set(to)
        cc = set(cc)
        bcc = set(bcc)

        if to.intersection(cc):
            raise ValidationError({"cc": _("Duplicate E-Mails in 'To' and 'CC'.")})
        if to.intersection(bcc):
            raise ValidationError({"bcc": _("Duplicate E-Mails in 'To' and 'BCC'.")})
        if cc.intersection(bcc):
            raise ValidationError({"bcc": _("Duplicate E-Mails in 'CC' and 'BCC'.")})
        return super().validate(data)

    class Meta(StepModelSerializer.Meta):
        model = EmailStep
        fields = StepModelSerializer.Meta.fields + ("template", "subject", "to", "_to", "cc", "_cc", "bcc", "_bcc")
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class EmailStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:emailstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = EmailStep
        fields = StepRepresentationSerializer.Meta.fields


class FinishStepModelSerializer(StepModelSerializer):
    def validate(self, data: dict) -> dict:
        data = super().validate(data)
        write_preserved_instance: bool | None = data.get(
            "write_preserved_instance", self.instance.write_preserved_instance if self.instance else None
        )
        workflow: Workflow | None = data.get("workflow", self.instance.workflow if self.instance else None)
        if write_preserved_instance and workflow and not workflow.preserve_instance:
            raise ValidationError(
                {
                    "write_preserved_instance": _(
                        "Cannot write preserved instance without picking the option to preserve it when creating the workflow."
                    )
                }
            )
        return data

    class Meta(StepModelSerializer.Meta):
        model = FinishStep
        fields = StepModelSerializer.Meta.fields + ("write_preserved_instance",)
        read_only_fields = StepModelSerializer.Meta.read_only_fields


class FinishStepRepresentationSerializer(StepRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:finishstep-detail")

    class Meta(StepRepresentationSerializer.Meta):
        model = FinishStep
        fields = StepRepresentationSerializer.Meta.fields



import json
from datetime import datetime
from io import BytesIO
from typing import Generator, Optional

import pandas as pd
import requests
from django.db.models import QuerySet
from dynamic_preferences.registries import global_preferences_registry
from pandas.tseries.offsets import BDay
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.io.backends import AbstractDataBackend, register
from wbcore.contrib.io.models import ImportCredential

from ..utils import get_timedelta_import_currency_fx_rates


@register("Currency FX Rates", provider_key="fixerio", save_data_in_import_source=False, passive_only=False)
class DataBackend(AbstractDataBackend):
    ENDPOINT = "https://data.fixer.io/api/"
    HEADERS = {"user-agent": "stainly", "Content-Type": "application/json", "Accept": "application/json"}

    def get_provider_id(self, obj: Currency) -> str:
        return obj.key

    def get_default_queryset(self) -> QuerySet[Currency]:
        return Currency.objects.all()

    def __init__(
        self,
        import_credential: Optional[ImportCredential] = None,
        base_currency: Optional[str] = None,
        **kwargs,
    ):
        if not import_credential or not import_credential.authentication_token:
            raise ValueError("Fixer IO backend needs a valid import credential object")
        if not base_currency:
            base_currency = global_preferences_registry.manager()["currency__default_currency"]
        self.authentication_token = import_credential.authentication_token
        self.base_currency = base_currency

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: Optional[list[str]] = None,
        **kwargs,
    ) -> Generator[tuple[str, BytesIO], None, None]:
        execution_date = execution_time.date()
        start = kwargs.get("start", execution_date - BDay(get_timedelta_import_currency_fx_rates() + 1))
        params = {"access_key": self.authentication_token, "base": self.base_currency}
        if (
            obj_external_ids and len(obj_external_ids) < Currency.objects.count()
        ):  # We add symbols only if the requests objects count is lower than the maximum expected currency
            params["symbols"] = ",".join(obj_external_ids)
        res = []
        for _date in pd.date_range(start, execution_date, freq="B"):
            r = requests.get(f'{self.ENDPOINT}/{_date.strftime("%Y-%m-%d")}?', params=params, headers=self.HEADERS)
            if r.status_code == requests.codes.ok:
                res_json = r.json()
                if res_json and res_json["success"]:
                    res.append(res_json)
        if res:
            content_file = BytesIO()
            content_file.write(json.dumps(res).encode())
            file_name = f"fixerio_currency_fx_rates_{start:%Y-%m-%d}_{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
            yield file_name, content_file



from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.documents.models import DocumentType


class DocumentTypeRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:documents:documenttyperepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:documents:documenttype-detail")

    class Meta:
        model = DocumentType
        fields = ("id", "_detail", "name", "computed_str")


class DocumentTypeModelSerializer(serializers.ModelSerializer):
    _parent = DocumentTypeRepresentationSerializer(source="parent")
    document_count = serializers.IntegerField(read_only=True)

    @serializers.register_resource()
    def list_of_children(self, instance, request, user):
        base_url = reverse("wbcore:documents:documenttype-list", request=request)
        return {"list_of_children": f"{base_url}?parent={instance.id}"}

    @serializers.register_resource()
    def list_of_documents(self, instance, request, user):
        base_url = reverse("wbcore:documents:document-list", request=request)
        if hasattr(instance, "document_count") and instance.document_count > 0:
            return {"list_of_documents": f"{base_url}?document_type={instance.id}"}
        return {}

    class Meta:
        model = DocumentType
        fields = (
            "id",
            "name",
            "document_count",
            "parent",
            "_parent",
            "_additional_resources",
        )



from rest_framework.request import Request
from rest_framework.reverse import reverse
from wbcore.configs.decorators import register_config


@register_config
def assigned_workflow_steps(request: Request) -> tuple[str, dict[str, str]]:
    return "workflow", {
        "endpoint": reverse("wbcore:workflow:processstep-assigned-list", request=request),
    }



from .document_types import DocumentTypeModelViewSet, DocumentTypeRepresentationViewSet
from .documents import DocumentModelViewSet, DocumentRepresentationViewSet
from .document_model_relationships import DocumentModelRelationshipModelViewSet
from .shareable_links import (
    DownloadShareableLinkView,
    ShareableLinkAccessModelViewSet,
    ShareableLinkAccessRepresentationViewSet,
    ShareableLinkModelViewSet,
    ShareableLinkRepresentationViewSet,
)



import logging
from datetime import datetime, timedelta

from celery import shared_task
from django.utils import timezone
from dynamic_preferences.registries import global_preferences_registry

from .models import User

logger = logging.getLogger(__name__)


@shared_task
def delete_unregistered_user_account(prune_user_account_before_datetime: datetime | None = None):
    if not prune_user_account_before_datetime:
        global_preferences = global_preferences_registry.manager()
        max_hours = global_preferences["authentication__hours_before_deleting_unregistered_account"]
        prune_user_account_before_datetime = timezone.now() - timedelta(hours=max_hours)

    for user in User.objects.filter(
        date_joined__lt=prune_user_account_before_datetime, is_register=False, is_active=False
    ):
        # we wrap the deletion routine into a try catch block to avoid encountering unexpected error which would prevent the proper progression of this task
        try:
            if user.profile:
                user.profile.delete()
            user.delete()
        except Exception as e:
            logger.warning(f"Couldn't delete unregistered user {user} because of error {e}")



from . import AbstractBackend


class IconBackend(AbstractBackend):
    """
    List of icons supported by the initial Stainly material
    """

    @classmethod
    @property
    def fallback_icon(cls):
        return "wb-icon-text-bg-color"

    ADD = "wb-icon-plus"
    APPROVE = "wb-icon-thumb-up"
    BANK = "wb-icon-bank"
    BIRTHDAY = "wb-icon-cake-1"
    BOOKMARK = "wb-icon-bookmark"
    CALENDAR = "wb-icon-calendar-1"
    CHART_AREA = "wb-icon-chart-area"
    CHART_BARS_HORIZONTAL = "wb-icon-chart-bars-1"
    CHART_BARS_VERTICAL = "wb-icon-chart-bars-4"
    CHART_BARS_VERTICAL_INCREASING = "wb-icon-chart-bars-increasing"
    CHART_LINE = "wb-icon-chart-line-2"
    CHART_PIE = "wb-icon-chart-pie"
    CHART_PYRAMID = "wb-icon-chart-piramid"
    CHART_SWITCHES = "wb-icon-chart-switches"
    CHART_WATERFALL = "wb-icon-chart-waterfall"
    CHART_MULTILINE = "wb-icon-chart-line-2"
    CLIPBOARD = "wb-icon-clipboard-2"
    CONFIRM = "wb-icon-circle-check-inverted"
    CONFIGURE = "wb-icon-configure"
    COPY = "wb-icon-copy-3"
    DASHBOARD = "wb-icon-dashboard"
    DATA_GRID = "wb-icon-data-grid-1"
    DATA_LIST = "wb-icon-data-list-6"
    DAY_OFF = "wb-icon-filters-off-2"
    DEAL = "wb-icon-handshake"
    DEAL_MONEY = "wb-icon-handshake-euro"
    DECISION_STEP = "wb-icon-circle-check-inverted"
    DELETE = "wb-icon-trashbin-2"
    DENY = "wb-icon-thumb-down"
    DISABLED = "wb-icon-disabled"
    DOCUMENT = "wb-icon-document-1"
    DOCUMENT_IN_PROGRESS = "wb-icon-document-gear"
    DOCUMENT_PRIVATE = "wb-icon-document-lock"
    DOLLAR = "wb-icon-dollar"
    DOWNLOAD = "wb-icon-triangle-down"
    EDIT = "wb-icon-edit-1"
    END = "wb-icon-dedent"
    ENUMERATION = "wb-icon-bulleted-list"
    EURO = "wb-icon-euro"
    EVENT = "wb-icon-calendar-3"
    EVENT_ACCEPTED = "wb-icon-event-accepted"
    EXPEND = "wb-icon-maximize-1"
    FAVORITE = "wb-icon-star"
    FEEDBACK = "wb-icon-envelope-open-person"
    FOLDERS_ADD = "wb-icon-folder-plus"
    FOLDERS_MONEY = "wb-icon-folder-euro"
    FOLDERS_OPEN = "wb-icon-folder"
    GENERATE_NEXT = "wb-icon-refresh-clock"
    GOAL = "wb-icon-calendar-1"
    GROUPS = "wb-icon-people"
    HOME = "wb-icon-home"
    HISTORY = "wb-icon-schedule"
    IGNORE = "wb-icon-eye-crossed"
    IMPORT_EXPORT = "wb-icon-trade"
    INFO = "wb-icon-circle-info"
    INVERSE = "wb-icon-arrows-swap-2"
    JOIN_STEP = "wb-icon-send"
    LINK = "wb-icon-link"
    LOCATION = "wb-icon-pin"
    LOCK = "wb-icon-lock-closed-2"
    LUNCH = "wb-icon-lunch"
    MAIL = "wb-icon-envelope"
    MAIL_MONEY = "wb-icon-envelope-open-euro"
    MAIL_OPEN = "wb-icon-envelope-open-1"
    MAIL_POSTED = "wb-icon-envelope-open-suitcase"
    MERGE = "wb-icon-copy"
    NEWSPAPER = "wb-icon-newspaper"
    NEXT = "wb-icon-triangles-right"
    NOTEBOOK = "wb-icon-notebook"
    OTHER = "wb-icon-data-list-5"
    PEOPLE = "wb-icon-people"
    PENDING = "wb-icon-hourglass-full"
    PERSON = "wb-icon-person"
    PERSON_ADD = "wb-icon-plus"
    PHONE = "wb-icon-phone"
    PHONE_ADD = "wb-icon-plus"
    PREVIOUS = "wb-icon-triangles-left"
    PUBLIC = "wb-icon-globe-1"
    QUESTION = "wb-icon-circle-question"
    REDO = "wb-icon-redo"
    REFRESH = "wb-icon-refresh-1"
    REGENERATE = "wb-icon-refresh-dots-gear"
    REJECT = "wb-icon-circle-x-inverted"
    REPEAT = "wb-icon-refresh-three"
    REPLACE = "wb-icon-arrows-replace"
    REVIEW = "wb-icon-edit-3"
    RUNNING = "wb-icon-hourglass-middle"
    SAVE = "wb-icon-floppy-disk"
    SCHEDULE = "wb-icon-clock-2"
    SCRIPT = "wb-icon-text-bg-color"
    SEND = "wb-icon-send"
    SEND_LATER = "wb-icon"
    SETTINGS = "wb-icon-gear-1"
    SHARE = "wb-icon-share-1"
    SHRINK = "wb-icon-minimize-1"
    SICK_LEAVE = "wb-icon-sick-leave-4"
    SPLIT = "wb-icon-arrows-swap-1"
    START = "wb-icon-step-forward-1"
    STATS = "wb-icon-chart-line-1"
    SUPERVISE = "wb-icon-people"
    SYNCHRONIZE = "wb-icon-refresh-dots"
    TABLE = "wb-icon-table"
    TIME_UP = "wb-icon-hourglass-empty"
    TRADE = "wb-icon-trade"
    TRAINING = "wb-icon-crosshair"
    UNDO = "wb-icon-undo"
    UNFILTER = "wb-icon-filters-off-3"
    UNFOLD = "wb-icon-triangle-down"
    UNLINK = "wb-icon-unlink"
    UNLOCK = "wb-icon-lock-open-2"
    UPLOAD = "wb-icon-triangle-up"
    VACATION = "wb-icon-suitcase-2"
    VIEW = "wb-icon-eye"
    WARNING = "wb-icon-circle-exclamation-inverted"
    WORK = "wb-icon-suitcase-1"


# class WBIcon(ChoiceEnum):
#     ACTION = "wb-icon-action"
#     ARROWS_SWAP_1 = "wb-icon-arrows-swap-1"
#     ARROWS_SWAP_2 = "wb-icon-arrows-swap-2"
#     ARROW_DOWN = "wb-icon-arrow-down"
#     ARROW_LEFT = "wb-icon-arrow-left"
#     ARROW_RIGHT = "wb-icon-arrow-right"
#     ARROW_UP = "wb-icon-arrow-up"
#     BANK = "wb-icon-bank"
#     BELL_EMPTY = "wb-icon-bell-inverted"
#     BELL_FULL = "wb-icon-bell"
#     BOLD = "wb-icon-bold"
#     BOOKMARK = "wb-icon-bookmark"
#     BOOKMARK_INVERTED = "wb-icon-bookmark-inverted"
#     BULLETED_LIST = "wb-icon-bulleted-list"
#     CAKE_1 = "wb-icon-cake-1"
#     CAKE_2 = "wb-icon-cake-2"
#     CAKE_3 = "wb-icon-cake-3"
#     CALENDAR_1 = "wb-icon-calendar-1"
#     CALENDAR_2 = "wb-icon-calendar-2"
#     CALENDAR_3 = "wb-icon-calendar-3"
#     CAMERA_1 = "wb-icon-camera-1"
#     CAMERA_2 = "wb-icon-camera-2"
#     CHART_BARS_1 = "wb-icon-chart-bars-1"
#     CHART_BARS_2 = "wb-icon-chart-bars-2"
#     CHART_BARS_3 = "wb-icon-chart-bars-3"
#     CHART_BARS_4 = "wb-icon-chart-bars-4"
#     CHART_LINE_2 = "wb-icon-chart-line-2"
#     CHART_PIE = "wb-icon-chart-pie"
#     CHART_PYRAMID = "wb-icon-chart-piramid"
#     CHART_SWITCHES = "wb-icon-chart-switches"
#     CIRCLE_CHECK = "wb-icon-circle-check"
#     CIRCLE_EXCLAMATION = "wb-icon-circle-exclamation"
#     CIRCLE_EXCLAMATION_INVERTED = "wb-icon-circle-exclamation-inverted"
#     CIRCLE_NO = "wb-icon-circle-x-inverted"
#     CIRCLE_QUESTION = "wb-icon-circle-question"
#     CIRCLE_QUESTION_INVERTED = "wb-icon-circle-question-inverted"
#     CIRCLE_X = "wb-icon-circle-x"
#     CIRCLE_YES = "wb-icon-circle-check-inverted"
#     CLIPBOARD_1 = "wb-icon-clipboard-1"
#     CLIPBOARD_2 = "wb-icon-clipboard-2"
#     CLIPBOARD_3 = "wb-icon-clipboard-3"
#     CLIPBOARD_4 = "wb-icon-clipboard-4"
#     CLIPBOARD_CHECKS_1 = "wb-icon-clipboard-checks-1"
#     CLIPBOARD_CHECKS_2 = "wb-icon-clipboard-checks-2"
#     CLOCK = "wb-icon-clock-2"
#     CLOCK_1 = "wb-icon-clock-1"
#     CLOCK_3 = "wb-icon-clock-3"
#     CLOCK_ALARM = "wb-icon-clock-alarm"
#     COG = "wb-icon-gear-6"
#     CONTACT_BOOK = "wb-icon-contact-book"
#     COPY_1 = "wb-icon-copy-1"
#     COPY_2 = "wb-icon-copy-2"
#     COPY_3 = "wb-icon-copy-3"
#     COPY_4 = "wb-icon-copy-4"
#     COPY_5 = "wb-icon-copy-5"
#     COPY_6 = "wb-icon-copy-6"
#     COPY_7 = "wb-icon-copy-7"
#     COPY_FAT_1 = "wb-icon-copy-fat-1"
#     COPY_FAT_2 = "wb-icon-copy-fat-2"
#     COPY_FAT_3 = "wb-icon-copy-fat-3"
#     COPY_FAT_4 = "wb-icon-copy-fat-4"
#     COPY_FAT_5 = "wb-icon-copy-fat-5"
#     CROSSHAIR = "wb-icon-crosshair"
#     DASHBOARD = "wb-icon-dashboard-2"
#     DASHBOARD_1 = "wb-icon-dashboard-1"
#     DATA = "wb-icon-data-grid-3"
#     DATA_GRID_1 = "wb-icon-data-grid-1"
#     DATA_GRID_2 = "wb-icon-data-grid-2"
#     DATA_GRID_4 = "wb-icon-data-grid-4"
#     DATA_GRID_5 = "wb-icon-data-grid-5"
#     DATA_LIST_2 = "wb-icon-data-list-2"
#     DATA_LIST_3 = "wb-icon-data-list-3"
#     DATA_LIST_4 = "wb-icon-data-list-4"
#     DATA_LIST_5 = "wb-icon-data-list-5"
#     DATA_LIST_6 = "wb-icon-data-list-6"
#     DATA_LIST_7 = "wb-icon-data-list-7"
#     DEDENT = "wb-icon-dedent"
#     DISABLED = "wb-icon-disabled"
#     DOCUMENT_1 = "wb-icon-document-1"
#     DOCUMENT_2 = "wb-icon-document-2"
#     DOCUMENT_BOOK = "wb-icon-document-book"
#     DOCUMENT_GEAR = "wb-icon-document-gear"
#     DOCUMENT_LOCK = "wb-icon-document-lock"
#     DOLLAR = "wb-icon-dollar"
#     EDIT_2 = "wb-icon-edit-2"
#     EDIT_3 = "wb-icon-edit-3"
#     ENVELOP = "wb-icon-envelope"
#     ENVELOPE_OPEN_1 = "wb-icon-envelope-open-1"
#     ENVELOPE_OPEN_2 = "wb-icon-envelope-open-2"
#     ENVELOPE_OPEN_DOLLAR = "wb-icon-envelope-open-dollar"
#     ENVELOP_EURO = "wb-icon-envelope-open-euro"
#     ENVELOP_LOCK = "wb-icon-envelope-open-suitcase"
#     ENVELOP_OPEN = "wb-icon-envelope-open-3"
#     ENVELOP_PERSON = "wb-icon-envelope-open-person"
#     EURO = "wb-icon-euro"
#     EYE = "wb-icon-eye"
#     EYE_CROSSED = "wb-icon-eye-crossed"
#     FEMALE = "wb-icon-female"
#     FILTERS_1 = "wb-icon-filters-1"
#     FILTERS_2 = "wb-icon-filters-2"
#     FILTERS_OFF_1 = "wb-icon-filters-off-1"
#     FILTERS_OFF_2 = "wb-icon-filters-off-2"
#     FILTERS_OFF_3 = "wb-icon-filters-off-3"
#     FLOPPY_DISK = "wb-icon-floppy-disk"
#     FLOPPY_DISK_CHECK = "wb-icon-floppy-disk-check"
#     FLOPPY_DISK_X = "wb-icon-floppy-disk-x"
#     FOLDER = "wb-icon-folder"
#     FOLDERS = "wb-icon-folders"
#     FOLDER_EURO = "wb-icon-folder-euro"
#     FOLDER_INVERTED = "wb-icon-folder-inverted"
#     FOLDER_PLUS = "wb-icon-folder-plus"
#     FOLDER_PLUS_INVERTED = "wb-icon-folder-plus-inverted"
#     FONT_COLOR = "wb-icon-font-color"
#     GEAR_1 = "wb-icon-gear-1"
#     GEAR_2 = "wb-icon-gear-2"
#     GEAR_3 = "wb-icon-gear-3"
#     GEAR_4 = "wb-icon-gear-4"
#     GEAR_5 = "wb-icon-gear-5"
#     GEAR_7 = "wb-icon-gear-7"
#     GLOBE = "wb-icon-globe-1"
#     GLOBE_2 = "wb-icon-globe-2"
#     HANDSHAKE = "wb-icon-handshake"
#     HANDS_EURO = "wb-icon-handshake-euro"
#     HOME = "wb-icon-home"
#     HOURGLASS_EMPTY = "wb-icon-hourglass-empty"
#     HOURGLASS_FULL = "wb-icon-hourglass-full"
#     HOURGLASS_MIDDLE = "wb-icon-hourglass-middle"
#     IMAGE = "wb-icon-image"
#     INDENT = "wb-icon-indent"
#     INFO = "wb-icon-circle-info"
#     ITALIC = "wb-icon-italic"
#     LINK = "wb-icon-link"
#     LOCK_CLOSED = "wb-icon-lock-closed-2"
#     LOCK_CLOSED_1 = "wb-icon-lock-closed-1"
#     LOCK_CLOSED_3 = "wb-icon-lock-closed-3"
#     LOCK_OPEN = "wb-icon-lock-open-2"
#     LOCK_OPEN_1 = "wb-icon-lock-open-1"
#     LOCK_OPEN_3 = "wb-icon-lock-open-3"
#     LUNCH = "wb-icon-lunch"
#     MALE = "wb-icon-male"
#     MAN_1 = "wb-icon-man-1"
#     MAN_2 = "wb-icon-man-2"
#     MAN_3 = "wb-icon-man-3"
#     MAN_4 = "wb-icon-man-4"
#     MAN_5 = "wb-icon-man-5"
#     MAN_6 = "wb-icon-man-6"
#     MAN_7 = "wb-icon-man-7"
#     MAXIMIZE = "wb-icon-maximize-3"
#     MAXIMIZE_1 = "wb-icon-maximize-1"
#     MAXIMIZE_2 = "wb-icon-maximize-2"
#     MENU = "wb-icon-data-list-1"
#     MINIMIZE = "wb-icon-minimize-4"
#     MINIMIZE_1 = "wb-icon-minimize-1"
#     MINIMIZE_2 = "wb-icon-minimize-2"
#     MINIMIZE_3 = "wb-icon-minimize-3"
#     NOTEBOOK = "wb-icon-notebook"
#     NOTEBOOK_INVERTED = "wb-icon-notebook-inverted"
#     NUMBERED_LIST = "wb-icon-numbered-list"
#     PAGE_BREAK = "wb-icon-page-break"
#     PENCIL = "wb-icon-edit-1"
#     PEOPLE = "wb-icon-people"
#     PERSON = "wb-icon-person"
#     PHONE = "wb-icon-phone"
#     PHONE_CALL = "wb-icon-phone-call"
#     PIN = "wb-icon-pin"
#     PIPETTE_EMPTY = "wb-icon-pipette-empty"
#     PIPETTE_FULL = "wb-icon-pipette-full"
#     PLUS = "wb-icon-plus"
#     QUOTE = "wb-icon-quote"
#     RECORDER = "wb-icon-recorder"
#     REDO = "wb-icon-redo"
#     REFRESH_1 = "wb-icon-refresh-1"
#     REFRESH_2 = "wb-icon-refresh-2"
#     REFRESH_24_ACW = "wb-icon-refresh-24-acw"
#     REFRESH_24_CW = "wb-icon-refresh-24-cw"
#     REFRESH_3_ACW = "wb-icon-refresh-3-acw"
#     REFRESH_3_CW = "wb-icon-refresh-3-cw"
#     REFRESH_CLOCK = "wb-icon-refresh-clock"
#     REFRESH_DOTS = "wb-icon-refresh-dots"
#     REFRESH_DOTS_24 = "wb-icon-refresh-dots-24"
#     REFRESH_DOTS_GEAR = "wb-icon-refresh-dots-gear"
#     REFRESH_THREE = "wb-icon-refresh-three"
#     REMOVE_FORMATTING = "wb-icon-remove-formatting"
#     REPLACE = "wb-icon-arrows-replace"
#     SCISSORS = "wb-icon-scissors"
#     SEARCH = "wb-icon-search"
#     SEND = "wb-icon-send"
#     SICK_LEAVE_1 = "wb-icon-sick-leave-1"
#     SICK_LEAVE_2 = "wb-icon-sick-leave-2"
#     SICK_LEAVE_3 = "wb-icon-sick-leave-3"
#     SICK_LEAVE_4 = "wb-icon-sick-leave-4"
#     SHARE_1 = "wb-icon-share-1"
#     SHARE_2 = "wb-icon-share-2"
#     SKYPE = "wb-icon-skype"
#     SORT_A_Z = "wb-icon-from-a-to-z"
#     SORT_Z_A = "wb-icon-from-z-to-a"
#     SOUND_OFF = "wb-icon-sound-off"
#     SOUND_ON = "wb-icon-sound-on"
#     SPEACH_BUBBLE = "wb-icon-speech-bubble"
#     SPELL_CHECK = "wb-icon-spell-check"
#     STAR = "wb-icon-star"
#     STATS = "wb-icon-chart-line-1"
#     STEP_BACK_2 = "wb-icon-step-back-2"
#     STEP_FORWARD_2 = "wb-icon-step-forward-2"
#     STEP_LEFT = "wb-icon-step-back-1"
#     STEP_RIGHT = "wb-icon-step-forward-1"
#     STRIKETHROUGH = "wb-icon-strikethrough"
#     SUBSCRIPT = "wb-icon-subscript"
#     SUITCASE_1 = "wb-icon-suitcase-1"
#     SUITCASE_2 = "wb-icon-suitcase-2"
#     SUPERSCRIPT = "wb-icon-superscript"
#     TABLE = "wb-icon-table"
#     TEXT_ALIGN_CENTER = "wb-icon-text-align-center"
#     TEXT_ALIGN_JUSTIFY = "wb-icon-text-align-justify"
#     TEXT_ALIGN_LEFT = "wb-icon-text-align-left"
#     TEXT_ALIGN_RIGHT = "wb-icon-text-align-right"
#     TEXT_BG_COLOR = "wb-icon-text-bg-color"
#     THUMBS_DOWN = "wb-icon-thumb-down"
#     THUMBS_UP = "wb-icon-thumb-up"
#     TRADE = "wb-icon-trade"
#     TRASH = "wb-icon-trashbin-2"
#     TRASHBIN_1 = "wb-icon-trashbin-1"
#     TRIANGLES_DOWN = "wb-icon-triangles-down"
#     TRIANGLES_LEFT = "wb-icon-triangles-left"
#     TRIANGLES_LEFT_STOP = "wb-icon-triangles-left-stop"
#     TRIANGLES_RIGHT = "wb-icon-triangles-right"
#     TRIANGLES_RIGHT_STOP = "wb-icon-triangles-right-stop"
#     TRIANGLES_UP = "wb-icon-triangles-up"
#     TRIANGLE_DOWN = "wb-icon-triangle-down"
#     TRIANGLE_LEFT = "wb-icon-triangle-left"
#     TRIANGLE_RIGHT = "wb-icon-triangle-right"
#     TRIANGLE_UP = "wb-icon-triangle-up"
#     UNDERLINE = "wb-icon-underline"
#     UNDO = "wb-icon-undo"
#     UNLINK = "wb-icon-unlink"
#     WOMAN_1 = "wb-icon-woman-1"
#     WOMAN_2 = "wb-icon-woman-2"
#     WOMAN_3 = "wb-icon-woman-3"
#     WOMAN_4 = "wb-icon-woman-4"
#     WOMAN_5 = "wb-icon-woman-5"
#     WOMAN_6 = "wb-icon-woman-6"
#     WOMAN_7 = "wb-icon-woman-7"
#     X = "wb-icon-x"



from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.workflow.models import Data, Step, Workflow


class WorkflowFilter(wb_filters.FilterSet):
    model = wb_filters.CharFilter(label=_("Attached Model"), method="filter_attached_model")
    step = wb_filters.ModelChoiceFilter(
        label=_("Associated Step"),
        queryset=Step.objects.all(),
        endpoint=Step.get_representation_endpoint(),
        value_key=Step.get_representation_value_key(),
        label_key=Step.get_representation_label_key(),
        method="filter_by_step",
    )
    # data = wb_filters.ModelChoiceFilter(
    #     label=_("Associated Data"),
    #     queryset=Data.objects.all(),
    #     endpoint=Data.get_representation_endpoint(),
    #     value_key=Data.get_representation_value_key(),
    #     label_key=Data.get_representation_label_key(),
    #     method="filter_by_data",
    # )

    def filter_attached_model(self, queryset: QuerySet[Workflow], name, value: str) -> QuerySet[Workflow]:
        if value:
            return queryset.filter(model__model=value)
        return queryset

    def filter_by_data(self, queryset: QuerySet[Workflow], name, value: Data | None) -> QuerySet[Workflow]:
        if value:
            return queryset.filter(attached_data=value)
        return queryset

    class Meta:
        model = Workflow
        fields = {
            "name": ["exact", "icontains"],
            "single_instance_execution": ["exact"],
            "preserve_instance": ["exact"],
            "status_field": ["exact", "icontains"],
        }



from django.utils.translation import gettext


def repeat(n: int, unit: str) -> str:
    """Creates a CSS repeat element, that makes the `unit` be repeated `n` times.

    Example:
        `repeat(5, Style.px(20))` results in the string `'repeat(5, 20px)'`

    Attributes:
        n (int): The amount of times the unit should be repeated
        unit (str): The unit that is going to be repeated. Should be generated by the unit helper methods, e.g. `Style.px(20)`

    Returns:
        A valid css string containing the information about repeated an element `n` times.
    """

    return f"repeat({n}, {unit})"


def repeat_field(n: int, field_name: str) -> str:
    """Creates a string repeating the `field_name` `n` times

    Example:
        `repeat_field(5, Style.px(20))` results in the string `'20px 20px 20px 20px 20px'`

    Args:
        n: An integer specifying how often the the `field_name` should be repeated
        field_name: A string, which should be repeated `n` times

    Returns:
        A string where the value of `field_name` is duplicated `n` times
    """
    return " ".join([field_name] * n)


def grid_definition(*units: str) -> str:
    """Concatanates the arguments specified as positional arguments with a space to create a valid css string to define a grid

    Arguments:
        *units: Positional string arguments that will be concatanated

    Returns:
        A valid css string of space seperated strings

    """
    return " ".join(units)


def split_list_into_grid_template_area_sublists(elem_list: list, column_number: int) -> list[list]:
    """Splits a list of elements into sublists of the provided column number length and adds empty grid cells at the end if necessary

    Args:
        elem_list (list): List of elements to split
        column_number (int): Number of columns in the grid. Must be above 0

    Returns:
        list: A list of lists of exactly the specified length

    Raises:
        ValueError: If column_number below 1 is provided
    """

    if column_number < 1:
        raise ValueError(gettext("Column number has to be at least 1!"))

    sublists = [elem_list]
    if column_number <= len(elem_list):
        # If we don't have enough elements in the list to split into the target column number we avoid the IndexError and instead populate the sublist with empty fields
        sublists = [elem_list[x : x + column_number] for x in range(0, len(elem_list), column_number)]

    for _ in range(column_number - len(sublists[-1])):
        sublists[-1].append(".")
    return sublists



from .contacts import TelephoneContactButtonConfig
from .entries import (
    CompanyModelButtonConfig,
    EntryModelButtonConfig,
    PersonModelButtonConfig,
)
from .relationships import EmployerEmployeeRelationshipButtonConfig



# Generated by Django 4.1 on 2022-10-04 06:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("authentication", "0004_token"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="metadata",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class StadiumTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Stadiums")

    def get_create_title(self) -> str:
        return _("Create Stadium")

    def get_instance_title(self) -> str:
        return _("Stadium")



from django.db.models import Q, QuerySet
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from wbcore import filters as wb_filters
from wbcore.contrib.authentication.models import Group, Permission, User
from wbcore.contrib.directory.models import EmailContact
from wbcore.contrib.workflow.models import (
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    StartStep,
    Step,
    Transition,
    UserStep,
    Workflow,
)
from wbcore.contrib.workflow.sites import workflow_site


class StepBaseFilter(wb_filters.FilterSet):
    code = wb_filters.NumberFilter(label=_("Code"), field_name="code", lookup_expr="exact")
    code__lte = wb_filters.NumberFilter(label=_("Code"), field_name="code", lookup_expr="lte")
    code__gte = wb_filters.NumberFilter(label=_("Code"), field_name="code", lookup_expr="gte")
    transition = wb_filters.ModelChoiceFilter(
        label=_("Associated Transition"),
        queryset=Transition.objects.all(),
        endpoint=Transition.get_representation_endpoint(),
        value_key=Transition.get_representation_value_key(),
        label_key=Transition.get_representation_label_key(),
        method="filter_by_transition",
    )

    def filter_by_transition(self, queryset: QuerySet[Step], name, value: Transition) -> QuerySet[Step]:
        if value:
            return queryset.filter(Q(outgoing_transitions=value) | Q(incoming_transitions=value))
        return queryset

    class Meta:
        model = Step
        fields = {
            "name": ["exact", "icontains"],
            "status": ["exact", "icontains"],
        }


class StepWorkflowFilter(StepBaseFilter):
    permission = wb_filters.ModelMultipleChoiceFilter(
        label=_("Permissions"),
        queryset=Permission.objects.all(),
        endpoint=Permission.get_representation_endpoint(),
        value_key=Permission.get_representation_value_key(),
        label_key=Permission.get_representation_label_key(),
    )
    step_type = wb_filters.MultipleChoiceFilter(label=_("Step Types"), choices=Step.StepType.choices)

    class Meta(StepBaseFilter.Meta):
        model = Step
        fields = StepBaseFilter.Meta.fields


class StepFilter(StepWorkflowFilter):
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )

    class Meta(StepWorkflowFilter.Meta):
        model = Step
        fields = StepWorkflowFilter.Meta.fields


class UserStepFilter(StepBaseFilter):
    permission = wb_filters.ModelMultipleChoiceFilter(
        label=_("Permissions"),
        queryset=Permission.objects.all(),
        endpoint=Permission.get_representation_endpoint(),
        value_key=Permission.get_representation_value_key(),
        label_key=Permission.get_representation_label_key(),
    )
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )
    assignee = wb_filters.ModelMultipleChoiceFilter(
        label=_("Assignees"),
        queryset=User.objects.all(),
        endpoint=User.get_representation_endpoint(),
        value_key=User.get_representation_value_key(),
        label_key=User.get_representation_label_key(),
    )

    group = wb_filters.ModelMultipleChoiceFilter(
        label=_("Groups"),
        queryset=Group.objects.all(),
        endpoint=Group.get_representation_endpoint(),
        value_key=Group.get_representation_value_key(),
        label_key=Group.get_representation_label_key(),
    )

    assignee_method = wb_filters.MultipleChoiceFilter(
        label=_("Assignee Methods"), choices=workflow_site.assignees_choices
    )

    # display = wb_filters.ModelMultipleChoiceFilter(
    #     label=_("Displays"),
    #     queryset=Display.objects.all(),
    #     endpoint=Display.get_representation_endpoint(),
    #     value_key=Display.get_representation_value_key(),
    #     label_key=Display.get_representation_label_key(),
    # )

    class Meta(StepBaseFilter.Meta):
        model = UserStep
        fields = StepBaseFilter.Meta.fields | {
            "notify_user": ["exact"],
        }


class DecisionStepFilter(StepBaseFilter):
    permission = wb_filters.ModelMultipleChoiceFilter(
        label=_("Permissions"),
        queryset=Permission.objects.all(),
        endpoint=Permission.get_representation_endpoint(),
        value_key=Permission.get_representation_value_key(),
        label_key=Permission.get_representation_label_key(),
    )
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )

    class Meta(StepBaseFilter.Meta):
        model = DecisionStep
        fields = StepBaseFilter.Meta.fields


class SplitStepFilter(DecisionStepFilter):
    class Meta(DecisionStepFilter.Meta):
        model = SplitStep
        fields = DecisionStepFilter.Meta.fields


class StartStepFilter(StepBaseFilter):
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )

    class Meta(StepBaseFilter.Meta):
        model = StartStep
        fields = StepBaseFilter.Meta.fields


class JoinStepFilter(DecisionStepFilter):
    class Meta(DecisionStepFilter.Meta):
        model = JoinStep
        fields = DecisionStepFilter.Meta.fields | {
            "wait_for_all": ["exact"],
        }


class ScriptStepFilter(DecisionStepFilter):
    class Meta(DecisionStepFilter.Meta):
        model = ScriptStep
        fields = DecisionStepFilter.Meta.fields | {
            "script": ["exact", "icontains"],
        }


class EmailStepFilter(DecisionStepFilter):
    template = wb_filters.CharFilter(label=_("Template"), method="filter_template_name")
    to = wb_filters.ModelMultipleChoiceFilter(
        label=pgettext_lazy("Email context", "To"),
        queryset=EmailContact.objects.all(),
        endpoint=EmailContact.get_representation_endpoint(),
        value_key=EmailContact.get_representation_value_key(),
        label_key=EmailContact.get_representation_label_key(),
    )
    cc = wb_filters.ModelMultipleChoiceFilter(
        label=_("CC"),
        queryset=EmailContact.objects.all(),
        endpoint=EmailContact.get_representation_endpoint(),
        value_key=EmailContact.get_representation_value_key(),
        label_key=EmailContact.get_representation_label_key(),
    )
    bcc = wb_filters.ModelMultipleChoiceFilter(
        label=_("BCC"),
        queryset=EmailContact.objects.all(),
        endpoint=EmailContact.get_representation_endpoint(),
        value_key=EmailContact.get_representation_value_key(),
        label_key=EmailContact.get_representation_label_key(),
    )

    def filter_template_name(self, queryset: QuerySet[EmailStep], name, value: str) -> QuerySet[EmailStep]:
        if value:
            return queryset.filter(template__endswith=value)
        return queryset

    class Meta(DecisionStepFilter.Meta):
        model = EmailStep
        fields = DecisionStepFilter.Meta.fields | {
            "subject": ["exact", "icontains"],
        }


class FinishStepFilter(DecisionStepFilter):
    class Meta(DecisionStepFilter.Meta):
        model = FinishStep
        fields = DecisionStepFilter.Meta.fields | {
            "write_preserved_instance": ["exact"],
        }



from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin


class PermissionTestModel(PermissionObjectModelMixin):
    class Meta(PermissionObjectModelMixin.Meta):
        managed = False



from configurations import values


class ImportExportBaseConfiguration:
    WBIMPORT_EXPORT_SFTPBACKEND_CLEAN_FILES = values.BooleanValue(False, environ_prefix=None)
    WBIMPORT_EXPORT_MAILBACKEND_SPAMSCORE = values.IntegerValue(None, environ_prefix=None)
    WBIMPORT_EXPORT_DEFAULT_EXPORT_PAGINATION_LIMIT = values.IntegerValue(200, environ_prefix=None)



from unittest.mock import patch

import pytest
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.authentication.models import Permission
from wbcore.contrib.directory.factories import EmailContactFactory, PersonFactory
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.models import EmailStep, ProcessStep, Step
from wbcore.contrib.workflow.sites import workflow_site
from wbcore.contrib.workflow.viewsets import (
    AssignedProcessStepModelViewSet,
    EmailStepModelViewSet,
    ProcessStepModelViewSet,
    WorkflowModelViewSet,
)
from wbcore.test.utils import get_data_from_factory, get_kwargs, get_or_create_superuser


@pytest.mark.django_db
class TestWorkflow:
    api_factory = APIRequestFactory()

    def test_get(self, workflow_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        workflow_factory.create_batch(3)
        vs = WorkflowModelViewSet.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    def test_retrieve(self, workflow_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        obj = workflow_factory()
        vs = WorkflowModelViewSet.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    def test_post(self, workflow_factory):
        obj = workflow_factory()
        super_user = get_or_create_superuser()
        data = get_data_from_factory(obj, WorkflowModelViewSet, superuser=super_user, delete=True)
        request = self.api_factory.post("", data=data)
        request.user = super_user
        kwargs = get_kwargs(obj, WorkflowModelViewSet, request)
        vs = WorkflowModelViewSet.as_view({"post": "create"})
        workflow_site.registered_model_classes_serializer_map[
            Person
        ] = "wbcore.contrib.directory.serializers.PersonModelSerializer"
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED

    def test_delete(self, workflow_factory):
        request = self.api_factory.delete("")
        request.user = get_or_create_superuser()
        obj = workflow_factory()
        kwargs = get_kwargs(obj, WorkflowModelViewSet, request)
        vs = WorkflowModelViewSet.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    def test_put(self, workflow_factory):
        old_obj = workflow_factory()
        new_obj = workflow_factory()
        user = get_or_create_superuser()
        data = get_data_from_factory(new_obj, WorkflowModelViewSet, superuser=user, delete=True)
        request = self.api_factory.put("", data=data)
        request.user = user
        vs = WorkflowModelViewSet.as_view({"put": "update"})
        workflow_site.registered_model_classes_serializer_map[
            Person
        ] = "wbcore.contrib.directory.serializers.PersonModelSerializer"
        response = vs(request, pk=old_obj.id)
        assert response.status_code == status.HTTP_200_OK

    def test_patch(self, workflow_factory):
        obj = workflow_factory()
        request = self.api_factory.patch("", data={"name": "New Name"})
        request.user = get_or_create_superuser()
        vs = WorkflowModelViewSet.as_view({"patch": "partial_update"})
        workflow_site.registered_model_classes_serializer_map[
            Person
        ] = "wbcore.contrib.directory.serializers.PersonModelSerializer"
        response = vs(request, pk=obj.id)
        assert response.status_code == status.HTTP_200_OK

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.start_workflow")
    def test_start_without_attached_instance(self, mock_start, workflow_factory, start_step_factory):
        workflow = workflow_factory(model=None, status_field=None, preserve_instance=False)
        start_step = start_step_factory(workflow=workflow)
        request = self.api_factory.patch("")
        request.GET = request.GET.copy()
        request.GET["step_id"] = start_step.pk
        user = get_or_create_superuser()
        request.user = user
        response = WorkflowModelViewSet().start(request, pk=workflow.pk)
        assert response.status_code == status.HTTP_200_OK
        assert mock_start.call_args.args == (start_step, None)

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.start_workflow")
    def test_start_with_attached_instance(self, mock_start, workflow_factory, start_step_factory):
        attached_person = PersonFactory()
        workflow = workflow_factory(preserve_instance=True)
        start_step = start_step_factory()
        request = self.api_factory.patch("")
        request.GET = request.GET.copy()
        request.GET["step_id"] = start_step.pk
        request.GET["instance_id"] = attached_person.pk
        user = get_or_create_superuser()
        request.user = user
        response = WorkflowModelViewSet().start(request, pk=workflow.pk)
        assert response.status_code == status.HTTP_200_OK
        assert mock_start.call_args.args == (
            start_step,
            attached_person,
        )


@pytest.mark.django_db
class TestEmailStep:
    api_factory = APIRequestFactory()

    def test_get(self, email_step_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        email_step_factory.create_batch(3)
        vs = EmailStepModelViewSet.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    def test_retrieve(self, email_step_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        obj = email_step_factory()
        vs = EmailStepModelViewSet.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    def test_post(self, email_step_factory):
        super_user = get_or_create_superuser()
        email_contact = EmailContactFactory()
        obj = email_step_factory(to=[email_contact])
        data = {
            "name": obj.name,
            "template": obj.template.open(mode="rb"),
            "step_type": obj.step_type,
            "subject": obj.subject,
            "to": obj.to.all().values_list("id", flat=True),
            "workflow": obj.workflow.id,
            "code": obj.code,
            "status": obj.status,
        }
        request = self.api_factory.post("", data=data)
        request.user = super_user
        EmailStep.objects.filter(pk=obj.pk).delete()
        kwargs = get_kwargs(obj, EmailStepModelViewSet, request)
        vs = EmailStepModelViewSet.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

        kwargs.pop("workflow_id")
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED

    def test_delete(self, email_step_factory):
        request = self.api_factory.delete("")
        request.user = get_or_create_superuser()
        obj = email_step_factory()
        kwargs = get_kwargs(obj, EmailStepModelViewSet, request)
        vs = EmailStepModelViewSet.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    def test_put(self, email_step_factory):
        email_contact1 = EmailContactFactory()
        email_contact2 = EmailContactFactory()
        old_obj = email_step_factory(to=[email_contact1])
        new_obj = email_step_factory(to=[email_contact2])
        user = get_or_create_superuser()
        data = {
            "name": new_obj.name,
            "template": new_obj.template.open(mode="rb"),
            "step_type": new_obj.step_type,
            "subject": new_obj.subject,
            "to": new_obj.to.all().values_list("id", flat=True),
            "workflow": new_obj.workflow.id,
            "code": new_obj.code,
            "status": new_obj.status,
        }
        request = self.api_factory.put("", data=data)
        request.user = user
        EmailStep.objects.filter(pk=new_obj.pk).delete()
        vs = EmailStepModelViewSet.as_view({"put": "update"})
        response = vs(request, pk=old_obj.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["name"] == new_obj.name
        assert not response.data["instance"]["name"] == old_obj.name

    def test_patch(self, email_step_factory):
        obj = email_step_factory()
        request = self.api_factory.patch("", data={"name": "New Name"})
        request.user = get_or_create_superuser()
        vs = EmailStepModelViewSet.as_view({"patch": "partial_update"})
        response = vs(request, pk=obj.id)
        assert response.status_code == status.HTTP_200_OK


@pytest.mark.django_db
class TestProcessStep:
    api_factory = APIRequestFactory()

    def test_get(self, process_step_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        process_step_factory.create_batch(3)
        vs = ProcessStepModelViewSet.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    def test_retrieve(self, process_step_factory):
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        obj = process_step_factory()
        vs = ProcessStepModelViewSet.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    def test_post(self, process_step_factory):
        obj = process_step_factory()
        super_user = get_or_create_superuser()
        data = get_data_from_factory(obj, ProcessStepModelViewSet, superuser=super_user, delete=True)
        request = self.api_factory.post("", data=data)
        request.user = super_user
        kwargs = get_kwargs(obj, ProcessStepModelViewSet, request)
        vs = ProcessStepModelViewSet.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    def test_delete(self, process_step_factory):
        request = self.api_factory.delete("")
        request.user = get_or_create_superuser()
        obj = process_step_factory()
        kwargs = get_kwargs(obj, ProcessStepModelViewSet, request)
        vs = ProcessStepModelViewSet.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    def test_patch(self, process_step_factory):
        obj = process_step_factory()
        request = self.api_factory.patch("", data={"name": "New Name"})
        request.user = get_or_create_superuser()
        vs = ProcessStepModelViewSet.as_view({"patch": "partial_update"})
        response = vs(request, pk=obj.id)
        assert response.status_code == status.HTTP_200_OK

    def test_get_queryset_superuser(self, process_step_factory):
        process_step1 = process_step_factory(permission=Permission.objects.first())
        process_step2 = process_step_factory()
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        mvs = ProcessStepModelViewSet(request=request, kwargs={})
        assert set(str(x.pk) for x in mvs.get_queryset()) == {process_step2.pk, process_step1.pk}

    def test_get_queryset_user_permission(self, process_step_factory):
        permission = Permission.objects.first()
        process_step1 = process_step_factory(permission=permission)
        process_step2 = process_step_factory()
        request = self.api_factory.get("")
        user = UserFactory()
        user.user_permissions.add(permission)
        request.user = user
        mvs = ProcessStepModelViewSet(request=request, kwargs={})
        assert set(str(x.pk) for x in mvs.get_queryset()) == {process_step2.pk, process_step1.pk}

    def test_get_queryset_group_permission(self, process_step_factory):
        permission = Permission.objects.first()
        process_step1 = process_step_factory(permission=permission)
        process_step2 = process_step_factory()
        request = self.api_factory.get("")
        group = GroupFactory(permissions=[permission])
        request.user = UserFactory(groups=[group])
        mvs = ProcessStepModelViewSet(request=request, kwargs={})
        assert set(str(x.pk) for x in mvs.get_queryset()) == {process_step1.pk, process_step2.pk}

    def test_get_queryset_no_permission(self, process_step_factory):
        process_step_factory(permission=Permission.objects.first())
        process_step2 = process_step_factory()
        request = self.api_factory.get("")
        request.user = UserFactory()
        mvs = ProcessStepModelViewSet(request=request, kwargs={})
        assert set(str(x.pk) for x in mvs.get_queryset()) == {process_step2.pk}

    @patch("wbcore.contrib.workflow.models.step.Step.start_next_step")
    def test_next(self, mock_next, process_step_factory, transition_factory):
        process_step = process_step_factory()
        transition = transition_factory()
        request = self.api_factory.get("")
        request.user = get_or_create_superuser()
        request.GET = request.GET.copy()
        request.GET["transition_id"] = transition.pk
        response = ProcessStepModelViewSet().next(request, pk=process_step.pk)
        assert response.status_code == status.HTTP_200_OK
        assert set(str(x.pk) for x in mock_next.call_args.args) == {process_step.pk, str(transition.pk)}

    @pytest.mark.parametrize(
        "bad_types",
        [
            Step.StepType.DECISIONSTEP,
            Step.StepType.EMAILSTEP,
            Step.StepType.FINISHSTEP,
            Step.StepType.JOINSTEP,
            Step.StepType.SCRIPTSTEP,
            Step.StepType.SPLITSTEP,
        ],
    )
    @pytest.mark.parametrize(
        "bad_states",
        [
            ProcessStep.StepState.CANCELED,
            ProcessStep.StepState.FAILED,
            ProcessStep.StepState.FINISHED,
            ProcessStep.StepState.WAITING,
        ],
    )
    def test_assigned_process_steps(self, process_step_factory, user_step_factory, bad_types, bad_states):
        request = self.api_factory.get("")
        group = GroupFactory()
        user = UserFactory(groups=[group], is_superuser=True)
        request.user = user
        step = user_step_factory(assignee_method=None)
        process_step1 = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        process_step2 = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, group=group)
        process_step_factory(step__step_type=bad_types, state=ProcessStep.StepState.ACTIVE, assignee=user)
        process_step_factory(step=step, state=bad_states, assignee=user)
        vs = AssignedProcessStepModelViewSet.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        id_set = {x["id"] for x in response.data.get("results")}
        assert id_set == {str(process_step1.pk), str(process_step2.pk)}
        assert response.status_code == status.HTTP_200_OK
